# SPDX-FileCopyrightText: 2021 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT
import datetime
from dataclasses import dataclass, field
from typing import List


@dataclass
class Sample:
    sample_uuid: str
    study_uuid: str = None
    sanger_sample_id: str = None
    location: str = None
    supplier_sample_name: str = None
    volume: str = None
    concentration: str = None
    public_name: str = None
    taxon_id: str = None
    common_name: str = None
    donor_id: str = None
    library_type: str = None
    country_of_origin: str = None
    sample_collection_date_utc: datetime = None
    cost_code: str = None
    genome_size: str = None
    accession_number: str = None
    sheared_femto_fragment_size: str = None
    post_spri_concentration: str = None
    post_spri_volume: str = None
    final_nano_drop_280: str = None
    final_nano_drop_230: str = None
    final_nano_drop: str = None
    shearing_and_qc_comments: str = None
    date_submitted_utc: datetime = None
    priority_level: str = None
    date_required_by: str = None
    reason_for_priority: str = None


@dataclass
class Update:
    uuid: str
    name: str
    value: str


@dataclass
class LabwareMessage:
    SUBJECT = None
    VERSION = 'latest'
    message_uuid: str
    message_create_date_utc: datetime = None


@dataclass
class CreateLabwareMessage(LabwareMessage):
    SUBJECT = 'create-labware'
    labware_type: str = None
    labware_uuid: str = None
    barcode: str = None
    samples: List[Sample] = field(default_factory=list)


@dataclass
class UpdateLabwareMessage(LabwareMessage):
    SUBJECT = 'update-labware'
    labware_updates: List[Update] = field(default_factory=list)
    sample_updates: List[Update] = field(default_factory=list)



# SPDX-FileCopyrightText: 2023 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

import typing
from abc import ABC, abstractmethod
from typing import Optional

from ._filterable import _Filterable

if typing.TYPE_CHECKING:
    from ..datasource_filter import DataSourceFilter
    from ..session import OperableSession


class Counter(_Filterable, ABC):
    """
    Counts number of results that match a filter
    """

    @abstractmethod
    def get_count(
        self,
        object_type: str,
        object_filters: Optional[DataSourceFilter] = None,
        session: Optional[OperableSession] = None
    ) -> int:
        """
        Counts the number of results that are matched by the (optional) filter
        """



# SPDX-FileCopyrightText: 2024 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

import os

from .defaults import Defaults
from ..core import (
    core_data_object
)
from ..goat import (
    GoatDataSource,
    create_goat_datasource
)


def goat(**kwargs) -> GoatDataSource:
    goat = create_goat_datasource(
        goat_url=os.getenv('GOAT_URL', Defaults.GOAT_URL)
        + os.getenv('GOAT_API_PATH', Defaults.GOAT_API_PATH)
    )
    core_data_object(goat)
    return goat



# SPDX-FileCopyrightText: 2023 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

import typing
from typing import Any, Union
from uuid import UUID

from prefect.client.schemas.objects import FlowRun

from .prefect_object import FlowRunObject
from ..core import DataSourceError
from ..core.core_converter import AsyncConverter

if typing.TYPE_CHECKING:
    from .prefect_datasource import PrefectDataSource


PrefectConverter = AsyncConverter[FlowRun, FlowRunObject]
"""
Converts from Prefect-native `FlowRun` to
`DataObject` instances.
"""
ObjectConverter = AsyncConverter[FlowRunObject, dict[str, Any]]
"""
Converts from `DataObject` to a "kwargs" `dict` for
`PrefectClient().create_flow_run_from_deployment()`
"""


class DefaultPrefectConverter(PrefectConverter):
    def __init__(
        self,
        prefect_ds: PrefectDataSource
    ) -> None:

        self.__prefect_ds = prefect_ds
        self.__names_cache: dict[str, tuple[str, str]] = {}

    async def async_convert(self, input_: FlowRun) -> FlowRunObject:
        attributes = await self.__get_attributes(input_)

        return self.__new_object(
            str(input_.id),
            attributes=attributes
        )

    async def __get_names(
        self,
        deployment_id: Union[str, UUID]
    ) -> tuple[str, str]:

        if deployment_id in self.__names_cache:
            return self.__names_cache[deployment_id]

        names = await self.__prefect_ds.get_names(
            deployment_id
        )
        self.__names_cache[deployment_id] = names
        return names

    async def __get_attributes(
        self,
        flow_run: FlowRun
    ) -> dict[str, Any]:

        dep_id = flow_run.deployment_id

        flow_name, deployment_name = await self.__get_names(dep_id)

        return {
            'name': flow_run.name,
            'flow_name': flow_name,
            'deployment_name': deployment_name,
            'state': flow_run.state_name,
            'tags': flow_run.tags,
            'idempotency_key': flow_run.idempotency_key,
            'parameters': flow_run.parameters
        }

    def __new_object(
        self,
        id_: str,
        attributes: dict[str, Any]
    ) -> FlowRunObject:

        return self.__prefect_ds.data_object_factory(
            'flow_run',
            id_,
            attributes=attributes
        )


class DefaultObjectConverter(ObjectConverter):

    __KEYS = (
        'parameters',
        'name',
        'tags',
        'idempotency_key'
    )

    def __init__(self, prefect_ds: PrefectDataSource) -> None:
        self.__ds = prefect_ds
        self.__dep_id_cache: dict[tuple[str, str], str] = {}

    async def async_convert(self, input_: FlowRunObject) -> dict[str, Any]:
        flow_name = input_.attributes.get('flow_name')
        dep_name = input_.attributes.get('deployment_name')

        self.__validate_names(flow_name, dep_name)

        deployment_id = await self.__get_deployment_id(flow_name, dep_name)

        return self.__to_kwargs(deployment_id, input_)

    def __validate_names(self, flow_name: str, dep_name: str) -> None:
        if not flow_name or not dep_name:
            detail = 'A name for both the flow and deployment are required'
            raise DataSourceError(
                title='Name Required',
                detail=detail,
                status_code=400
            )

    async def __get_deployment_id(
        self,
        flow_name: str,
        deployment_name: str
    ) -> Union[str, UUID]:

        __k = (flow_name, deployment_name)

        if __k in self.__dep_id_cache:
            return self.__dep_id_cache[__k]

        __v = await self.__ds.get_deployment_id(
            flow_name,
            deployment_name
        )
        self.__dep_id_cache[__k] = __v
        return __v

    def __to_kwargs(
        self,
        deployment_id: Union[str, UUID],
        obj: FlowRunObject
    ) -> dict[str, Any]:

        attributes = self.__get_relevant_attributes(obj)

        return {
            **attributes,
            'deployment_id': deployment_id
        }

    def __get_relevant_attributes(
        self,
        obj: FlowRunObject
    ) -> dict[str, Any]:

        return {
            k: obj.attributes.get(k)
            for k in self.__KEYS
        }



# SPDX-FileCopyrightText: 2023 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

import typing
from abc import ABC, abstractmethod
from typing import Any, Callable, Dict, List, Optional, Tuple

from cachetools.func import ttl_cache

from .attribute_metadata import (
    AttributeMetadata,
    DefaultAttributeMetadata
)
from .data_object import DataDict
from .datasource_error import DataSourceError, NoDataObjectFactoryError
from .factory import DataObjectFactory
from .operator import AllOperatorType
from .session import DataSourceSession

if typing.TYPE_CHECKING:
    from .session import OperableSession


DataId = str
DataSourceUpdate = Tuple[DataId, DataDict]
DataSourceConfig = Dict[str, Any]
UserIdGetter = Callable[[], str]


class DataSource(ABC):
    """
    The central class for managing operations on heterogeneous data sources.
    """

    DEFAULT_PAGE_SIZE = 20

    def __init__(
            self,
            config: DataSourceConfig,
            *args,
            expected: List[str] | None = None,
            attribute_metadata: type[AttributeMetadata] = DefaultAttributeMetadata,
            **kwargs):
        self.__data_object_factory: Optional[DataObjectFactory] = None
        self.__attribute_metadata = attribute_metadata
        self.__validate_config(config, expected)
        for k, v in config.items():
            setattr(self, k, v)

    @property
    @abstractmethod
    def supported_types(self) -> List[str]:
        """
        The list of types of DataObject supported by this DataSource instance.

        This can either be a static list, or dynamically generated.
        """

    def get_session(self) -> OperableSession:
        """
        Gets a `DataSourceSession` instance. This behaviour is
        `DataSource`-specific, and may be overriden.
        """

        return DataSourceSession(self)

    def __validate_config(
        self,
        config: DataSourceConfig,
        expected: List[str] | None,
    ):
        if expected is None:
            return
        for k in expected:
            if k not in config:
                raise DataSourceError(
                    title='Incorrect configuration',
                    detail=f'{k} missing in config dict'
                )

    def get_page_size(self) -> int:
        return getattr(self, 'page_size', self.DEFAULT_PAGE_SIZE)

    def get_attribute_types(self, object_type: str) -> dict[str, str]:
        """
        DEPRECATED - use the `DataSource().attribute_types` property
        instead.

        The types (str, int, etc) of the attributes of an object_type.

        This can either be a static list, or dynamically generated.
        """

        return self.attribute_types[object_type]

    @property
    def attribute_types(self) -> dict[str, dict[str, str]]:
        """
        The `dict` mapping supported types to the (python) types of
        their attributes.

        This can either be a static `dict`, or dynamically generated.
        """

        return {}

    @property
    @ttl_cache(ttl=86400)
    def attribute_metadata(self) -> dict[str, dict[str, dict[str, Optional[str | bool]]]]:
        """
        Information about the attributes:
            python_type
            display_name
            available_on_relationships
        """
        ret: dict[str, dict[str, dict[str, Any]]] = {}
        am = self.__attribute_metadata()
        for object_type, attribute in self.attribute_types.items():
            ret[object_type] = {}
            for attribute_name, attribute_type in attribute.items():
                ret[object_type][attribute_name] = {
                    'python_type': attribute_type,
                    'display_name': am.get_display_name(object_type, attribute_name),
                    'description': am.get_description(object_type, attribute_name),
                    'cardinality': am.get_cardinality(object_type, attribute_name),
                    'authoritative': am.is_authoritative(object_type, attribute_name),
                    'source': am.get_source(object_type, attribute_name),
                    'acts_as': am.get_acts_as(object_type, attribute_name),
                    'available_on_relationships':
                        am.is_available_on_relationships(
                            object_type,
                            attribute_name)
                }
        return ret

    def get_attribute_metadata_by_name(self, obj_type: str, field_name: str) -> Any:
        """
        Get attribute_metadata by name, or return `None` if the field does not exist.
        """
        # Split by dots to allow for nested fields
        field_names = field_name.split('.')
        current_obj_type = obj_type
        for name in field_names:
            # It may be an attribute of the current object type
            if name in self.attribute_metadata[current_obj_type]:
                return self.attribute_metadata[current_obj_type][name]
            # ...or a to_one relation
            if current_obj_type in self.relationship_config and \
                    self.relationship_config[current_obj_type].to_one is not None:
                if name in self.relationship_config[current_obj_type].to_one:
                    current_obj_type = self.relationship_config[current_obj_type].to_one[name]
                    continue
            return None
        return None

    @property
    def data_object_factory(self) -> DataObjectFactory:
        """A callable that returns a new DataObject for the given type."""

        if self.__data_object_factory is None:
            raise NoDataObjectFactoryError(
                'The `data_object_factory` setter must be called before a '
                'DataSource instance can be used. The standard way to do '
                'this is to collect all of the DataSource instances, and '
                'provide them to the `core_data_object` function.'
            )

        return self.__data_object_factory

    @data_object_factory.setter
    def data_object_factory(
        self,
        data_object_factory: DataObjectFactory
    ) -> None:
        """Sets the factory for creating new DataObject instances"""

        self.__data_object_factory = data_object_factory


class OperableDataSource(DataSource, AllOperatorType):
    """A type hint. For inheriting, use DataSource"""



# SPDX-FileCopyrightText: 2023 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

import typing
from abc import ABC, abstractmethod
from typing import Iterable, Optional

from ._filterable import _Filterable
from ..datasource_error import DataSourceError
from ..datasource_filter import DataSourceFilter

if typing.TYPE_CHECKING:
    from ..data_object import DataObject
    from ..requested_fields import ReqFieldsTree
    from ..session import OperableSession


class Cursor(_Filterable, ABC):
    """
    Implements cursor pagination.
    """

    @property
    def candidate_keys(self) -> dict[str, list[str]]:
        """
        The names of the fields that form the
        candidate_key for cursor pagination on
        an `object_type`.
        """

        return {
            k: ['id']
            for k in self.supported_types
        }

    @property
    @abstractmethod
    def supported_types(self) -> list[str]:
        pass

    @abstractmethod
    def get_cursor_page(
        self,
        object_type: str,
        page_size: Optional[int] = None,
        object_filters: Optional[DataSourceFilter] = None,
        search_after: list[str] | None = None,
        session: Optional[OperableSession] = None,
        requested_fields: list[str] | None = None,
        requested_tree: ReqFieldsTree | None = None,
    ) -> tuple[Iterable[DataObject], list[str] | None]:
        """
        Gets an `Iterable` of `DataObject` instances of the given
        `object_type`, of length given by `page_size`, from either:

        - the start (`if search_after is None`)
        - `search_after` (`if search_after is not None` i.e. an ID)

        Also returns the next `search_after` term.
        """

    def update_cursor_filters(
        self,
        search_after: list[str] | None,
        object_filters: DataSourceFilter | None
    ) -> DataSourceFilter | None:

        if search_after is None:
            return object_filters

        if len(search_after) != 1:
            raise DataSourceError(
                'Bad Cursor',
                'The `search_after` argument must specify an ID only.',
                400
            )

        if object_filters is None:
            return DataSourceFilter(
                and_={
                    'id': {
                        'gt': {
                            'value': search_after[0]
                        }
                    }
                }
            )

        if object_filters.and_ is None:
            object_filters.and_ = {}

        object_filters.and_['id'] = {
            'gt': {
                'value': search_after[0]
            }
        }
        return object_filters

    def can_use_cursor(
        self,
        object_type: str,
        object_filters: DataSourceFilter | None,
    ) -> bool:

        if object_filters is None or object_filters.and_ is None:
            return True

        keys = self.candidate_keys[object_type]

        return not any(
            True for k in keys
            if k in object_filters.and_
        )

    def _get_list_by_cursor(
        self,
        object_type: str,
        object_filters: Optional[DataSourceFilter] = None,
        session: Optional[OperableSession] = None,
        **kwargs,
    ) -> Iterable[DataObject]:
        """
        A usable implementation for `ListGetter.get_list()`
        using `Cursor.get_cursor_page()` internally.
        """

        search_after = None
        page_size = self.get_page_size()

        while True:
            fetched, search_after = self.get_cursor_page(
                object_type,
                page_size=page_size,
                object_filters=object_filters,
                search_after=search_after,
                session=session,
                **kwargs,
            )
            objs = list(fetched)

            yield from objs

            if len(objs) < page_size:
                return



# SPDX-FileCopyrightText: 2024 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

import string
from typing import Iterable, Optional

from .converter import DummyTransfer
from ..core import HttpClient


class DummyClient(HttpClient):
    """
    Simulates dummy objects retrieval from a remote API.
    """

    def __init__(
        self,
    ) -> None:
        super().__init__()

    def get_detail(
        self,
        object_type: str,
        object_ids: Iterable[str]
    ) -> Optional[DummyTransfer]:
        """
        Gets a list of Dummy transfers for the objects of specified
        `object_type` and `object_id`, or returns None if not found.
        """
        return self.__get_detail_data(object_type, object_ids)

    def get_list(
        self,
        object_type: str
    ) -> Optional[DummyTransfer]:
        return self.__get_detail_data(object_type, range(10000))

    def __get_detail_data(
        self,
        object_type: str,
        object_ids: Iterable[str]
    ) -> Optional[DummyTransfer]:
        if object_type == 'category':
            return self.__get_detail_category(object_type, object_ids)
        return self.__get_detail_record(object_type, object_ids)

    def __get_detail_record(
        self,
        object_type: str,
        object_ids: Iterable[str]
    ) -> Optional[DummyTransfer]:
        """
        Gets a list of dummy objects
        """
        ret = []
        for object_id in object_ids:
            if object_id < 10000:
                ret.append({
                    'id': object_id,
                    'little_string': ['a', 'b', 'c'][int(object_id) % 3],
                    'big_string': string.ascii_letters[int(object_id) % 52],
                    'int': object_id,
                    'bool': int(object_id) % 2 == 0,
                    'date': f'2024-01-{(int(object_id) % 28) + 1:02d}',
                    'type': object_type,
                    'category': ['cat1', 'cat2', 'cat3', 'cat4'][int(object_id) % 4],
                })
            else:
                ret.append(None)
        return ret

    def __get_detail_category(
        self,
        object_type: str,
        object_ids: Iterable[str]
    ) -> Optional[DummyTransfer]:
        """
        Gets a list of dummy objects
        """
        ret = []
        for object_id in object_ids:
            ret.append({
                'id': object_id,
                'name': object_id.upper(),
                'type': object_type
            })
        return ret



# SPDX-FileCopyrightText: 2024 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from functools import cache
from typing import Any, Dict, Iterable, Optional

from dateutil import parser as dateutil_parser

import requests

from ..core import (
    DataObject,
    DataSource,
    DataSourceError,
    DataSourceFilter
)
from ..core.operator import (
    DetailGetter,
    ListGetter
)


class JsonDataSource(
    DataSource,

    # the supported operators
    DetailGetter,
    ListGetter
):
    """
    A `DataSource` that connects to a remote JSON data file
    """

    def __init__(
        self,
        config: Dict
    ) -> None:
        super().__init__(
            config,
            expected=['uri', 'type', 'id_attribute', 'mappings']
        )
        self._raw_data = self._load_json()
        self._keyed_by_id = {
            v[self.id_attribute]: v
            for v in self._raw_data
            if self.id_attribute in v
        }

    def _load_json(self):
        response = requests.get(self.uri)
        if response.status_code != 200:
            raise DataSourceError()
        data = response.json()
        return data

    def __map_entry(self, entry: Dict):
        ret = {}
        for mapping_key, mapping_value in self.mappings.items():
            if mapping_value['heading'] != self.id_attribute:
                subentry = entry
                for level in mapping_value['heading'].split('.'):
                    if subentry is not None:
                        subentry = subentry.get(level)
                ret[mapping_key] = self.__parse_date(mapping_key, subentry)
        return ret

    def __create_data_object(self, entry: Dict):
        return self.data_object_factory(
            self.type,
            entry.get(self.id_attribute),
            attributes=self.__map_entry(entry)
        )

    def __parse_date(self, attribute_name: str, value: Any):
        if self.mappings[attribute_name]['type'] == 'datetime' and value is not None:
            return dateutil_parser.parse(value)
        if self.mappings[attribute_name]['type'] == 'str' and value is not None:
            return str(value)
        return value

    @property
    @cache
    def attribute_types(self) -> dict[str, dict[str, str]]:
        return {
            self.type: {
                k: v['type']
                for k, v in self.mappings.items()
            }
        }

    @property
    @cache
    def supported_types(self) -> list[str]:
        return list(
            self.attribute_types.keys()
        )

    def get_by_id(
        self,
        object_type: str,
        object_ids: Iterable[str],
        **kwargs,
    ) -> Iterable[Optional[DataObject]]:
        if object_type not in self.supported_types:
            raise DataSourceError(f'{object_type} is not supported')

        return (
            self.__create_data_object(self._keyed_by_id[object_id])
            if self._keyed_by_id[object_id] is not None else None
            for object_id in object_ids
        )

    def get_list(
        self,
        object_type: str,
        object_filters: Optional[DataSourceFilter] = None
    ) -> Iterable[DataObject]:
        """
        Gets an Iterable of DataObject instances of the given
        type, according to the given DataSourceFilter.
        """
        if object_filters is not None:
            raise DataSourceError('Filtering is not supported on JsonDataSource')
        if object_type not in self.supported_types:
            raise DataSourceError(f'{object_type} is not supported')
        for entry in self._raw_data:
            id_ = entry.get(self.id_attribute)
            if id_ is not None and id_ != '':
                yield self.__create_data_object(entry)



# SPDX-FileCopyrightText: 2022 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

import urllib3  # noqa
from .main import Interface  # noqa

urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)



# SPDX-FileCopyrightText: 2023 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from ...core import DataSourceError


class RelataionshipHopsParser:
    """
    Parses the relationship hops from a related URL.
    """

    def __init__(self, url_suffix: str) -> None:
        self.__suffix = url_suffix

    @property
    def relationship_hops(self) -> list[str]:
        hops = self.__suffix.split('/')
        self.__validate_all_hops(hops)
        return hops

    def __validate_all_hops(self, hops: list[str]) -> None:
        for hop in hops:
            self.__validate_hop(hop)

    def __validate_hop(self, hop: str) -> None:
        if not self.__hop_is_valid(hop):
            detail = (
                'The given recursive relationship '
                'path is invalid. Please check the '
                'provided information, and try again.'
            )
            raise DataSourceError(
                title='Invalid Relationship',
                detail=detail,
                status_code=400
            )

    def __hop_is_valid(self, hop: str) -> None:
        return hop and not hop.isspace()



# SPDX-FileCopyrightText: 2024 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

from datetime import datetime, timedelta
from typing import Any, Callable, Optional
from urllib.parse import urlencode

from requests.auth import HTTPBasicAuth

from .models import ModelClass, ModelTuple, create_models
from ..session import SessionFactory, create_session_factory
from ...api_base.auth import (
    AuthBlueprint,
    AuthManager,
    OidcConfig,
    StateNotFoundError,
    env_oidc_config,
)
from ...api_base.auth.abc import AuthorisationManager
from ...api_base.misc import AuthContext
from ...core import HttpClient


class DbAuthManager(AuthManager):
    """
    Implements `AuthManager` using SQLAlchemy and a database to achieve an OIDC workflow.

    This class provides database-backed authentication management with OpenID Connect (OIDC)
    integration. It handles the complete OIDC flow including state management, token exchange,
    user profile creation, authentication, and token revocation.

    The manager supports:
    - OIDC authorization code flow
    - Database storage of users, tokens, and state
    - Automatic cleanup of expired tokens and states
    - User profile creation from OIDC claims
    - Token-based authentication and authorization
    """

    def __init__(
        self,
        oidc_config: OidcConfig,
        session_factory: SessionFactory,
        model_tuple: ModelTuple,
        state_delete_delta: timedelta,
        oidc_id_target: str,
        oidc_ext_mapping: dict[str, str],
        authorisation_manager: AuthorisationManager,
    ) -> None:
        """
        Initialize the database authentication manager.

        Args:
            oidc_config: OIDC configuration containing URLs, client credentials, etc.
            session_factory: Factory function for creating database sessions
            model_tuple: Container with the SQLAlchemy model classes for auth tables
            state_delete_delta: Time delta after which OIDC state entries are cleaned up
            oidc_id_target: Field name in OIDC user info to use as the user identifier
            oidc_ext_mapping: Mapping from OIDC field names to local user attribute names
        """
        self.__config = oidc_config
        self.__session_factory = session_factory
        self.__models = model_tuple
        self.__state_delta = state_delete_delta
        self.__oidc_id_target = oidc_id_target
        self.__ext_map = oidc_ext_mapping
        self.__authorisation_manager = authorisation_manager

    def login(self) -> dict[str, str]:
        """
        Initiate the OIDC login flow by generating a new state and returning the authorization URL.

        This method performs cleanup of old states and expired tokens before creating a new
        login session. It generates a unique state parameter for CSRF protection and returns
        the complete authorization URL the user should be redirected to.

        Returns:
            dict: Dictionary containing the 'loginUrl' key with the complete authorization URL
        """
        self.__cleanup_before_login()
        state_uuid = self.__create_new_state()

        return self.__format_login_url(state_uuid)

    def get_token_from_callback(self, state: str, code: str) -> dict[str, str]:
        """
        Exchange the authorization code for an access token after OIDC callback.

        This method validates the state parameter to prevent CSRF attacks and then
        exchanges the authorization code for an access token by posting to the
        OIDC token endpoint.

        Args:
            state: The state parameter returned from the OIDC provider
            code: The authorization code returned from the OIDC provider

        Returns:
            dict: Token response from the OIDC provider, typically containing
                 access_token, token_type, expires_in, etc.

        Raises:
            StateNotFoundError: If the provided state parameter is not found or expired
        """
        self.__check_state_exists(state)

        return self.__post_to_token_url(code)

    def create_user_profile(self, token: str) -> dict[str, Any]:
        """
        Create or update a user profile using the provided access token.

        This method fetches user information from the OIDC provider using the access token,
        creates or updates the user record in the database, stores the token, and returns
        a complete user profile including OIDC ID and any extended user information.

        Args:
            token: Access token to use for fetching user information

        Returns:
            dict: Complete user profile containing:
                - oidc_id: The user's OIDC identifier
                - Any additional user attributes from the database
                - Any token-related extra information
        """
        oidc_id, oidc_ext = self.__get_oidc_details(token)
        user_id, user_ext = self.__get_or_create_user(oidc_id, oidc_ext)
        token_extra = self.__get_or_create_token(token, user_id)

        return {**token_extra, **user_ext, 'oidc_id': oidc_id}

    def authenticate(
        self,
        ctx: AuthContext,
        token: str,
    ) -> None:
        """
        Authenticate a user using the provided token and populate the auth context.

        This method looks up the user associated with the token and populates the
        authentication context with the user ID and role names if the token is valid.
        If the token is not found or invalid, the context remains unchanged.

        Args:
            ctx: Authentication context to populate with user information
            token: Access token to authenticate with
        """
        user_id, role_names, memberships = self.__get_user_details_for_token(token)
        if user_id is not None:
            ctx.user_id = user_id
            ctx.roles = role_names
            ctx.memberships = memberships

    def revoke_token(self, token: str) -> None:
        """
        Revoke an access token both locally and at the OIDC provider.

        This method deletes the token from the local database and then calls the
        OIDC provider's token revocation endpoint to invalidate the token.

        Args:
            token: Access token to revoke
        """
        self.__delete_token(token)
        self.__post_revoke(token)

    def __map_oidc_extra(self, json_return: dict[str, Any]) -> dict[str, Any]:
        """
        Map OIDC user information fields to local user attribute names.

        Uses the oidc_ext_mapping configuration to transform field names from the
        OIDC provider response to the expected local attribute names.

        Args:
            json_return: User information response from OIDC provider

        Returns:
            dict: Mapped user attributes with local field names
        """
        return {v: json_return.get(k) for k, v in self.__ext_map.items()}

    def __post_revoke(self, token: str) -> None:
        """
        Revoke a token at the OIDC provider's revocation endpoint.

        Makes a POST request to the OIDC provider's token revocation endpoint
        with the token and client credentials.

        Args:
            token: Access token to revoke

        Raises:
            requests.HTTPError: If the revocation request fails
        """

        client = HttpClient()

        session = client.get_session()

        r = session.post(
            self.__config.revoke_url,
            data={
                'token': token,
                'client_id': self.__config.client_id,
                'client_secret': self.__config.client_secret,
            },
        )
        r.raise_for_status()

    def __get_oidc_details(self, token: str) -> tuple[str, dict[str, Any]]:
        """
        Fetch user details from the OIDC provider using an access token.

        Makes a request to the OIDC user info endpoint with the Bearer token
        and extracts the user ID and additional mapped attributes.

        Args:
            token: Access token to use for the user info request

        Returns:
            tuple: A tuple containing:
                - str: User's OIDC identifier
                - dict: Mapped additional user attributes

        Raises:
            requests.HTTPError: If the user info request fails
        """

        client = HttpClient()

        session = client.get_session()

        headers = {'Authorization': f'Bearer {token}'}

        r = session.get(self.__config.user_info_url, headers=headers)
        r.raise_for_status()

        json_return = r.json()

        return (json_return[self.__oidc_id_target], self.__map_oidc_extra(json_return))

    def __get_user_details_for_token(
        self, token: str
    ) -> tuple[Optional[str], list[str], list[str]]:
        """
        Retrieve user details associated with a token from the database.

        Looks up the token in the database and returns the associated user ID
        and role names. Returns None and empty list if token is not found.

        Args:
            token: Access token to look up

        Returns:
            tuple: A tuple containing:
                - Optional[str]: User ID if token exists, None otherwise
                - list[str]: List of role names for the user, empty if no token
                - list[str]: List of membership identifiers for the user, empty if no token
        """
        token_model = self.__models.token_class

        with self.__session_factory() as sess:
            instance = token_model.get(sess, token)

            if instance is None:
                return None, [], []
            else:
                user = instance.user

                if self.__authorisation_manager:
                    memberships = self.__authorisation_manager.get_user_memberships(user)
                else:
                    memberships = []

                return str(user.id), user.role_names, memberships

    def __get_or_create_user(
        self, oidc_id: str, oidc_ext: dict[str, Any]
    ) -> tuple[int, dict[str, Any]]:
        """
        Get an existing user or create a new one using OIDC information.

        Looks up a user by their OIDC ID and creates a new user if one doesn't exist.
        Returns the user ID and any extended user information.

        Args:
            oidc_id: The user's OIDC identifier
            oidc_ext: Additional user attributes from OIDC mapping

        Returns:
            tuple: A tuple containing:
                - int: User's database ID
                - dict: Extended user information from the user model
        """
        user_model = self.__models.user_class

        with self.__session_factory() as sess:
            user = user_model.get_or_create(sess, oidc_id, **oidc_ext)
            userinfo_ext = self.__get_user_ext(user)

            return user.id, userinfo_ext

    def __get_user_ext(self, user: Any) -> dict[str, Any]:
        """
        Extract extended user information from a user model instance.

        If the user model defines a `get_userinfo_ext()` method, calls it to
        get additional user information. Otherwise returns an empty dictionary.

        Args:
            user: User model instance

        Returns:
            dict: Extended user information, empty if method not defined
        """
        if self.__user_model_defines_ext(user):
            return user.get_userinfo_ext()
        else:
            return {}

    def __user_model_defines_ext(self, user: Any) -> bool:
        """
        Check if the user model defines a method for extended user information.

        Returns `True` if a `get_userinfo_ext()` method is
        defined on the mixin for `User`.

        Args:
            user: User model instance to check

        Returns:
            bool: True if the user model has a get_userinfo_ext method, False otherwise
        """
        return hasattr(user, 'get_userinfo_ext') and callable(user.get_userinfo_ext)

    def __get_or_create_token(self, token: str, user_id: int) -> str:
        """
        Store a token in the database or update existing token record.

        Creates or updates a token record associated with the given user ID.

        Args:
            token: Access token to store
            user_id: ID of the user associated with the token

        Returns:
            str: Any additional token information returned by the model
        """
        token_model = self.__models.token_class

        with self.__session_factory() as sess:
            return token_model.get_or_create(sess, token, user_id)

    def __delete_token(self, token: str) -> str:
        """
        Delete a token from the database.

        Removes the token record from the database.

        Args:
            token: Access token to delete

        Returns:
            str: Result of the delete operation from the model
        """
        token_model = self.__models.token_class

        with self.__session_factory() as sess:
            return token_model.delete(sess, token)

    def __check_state_exists(self, state_uuid: str) -> None:
        """
        Verify that an OIDC state parameter exists in the database.

        Checks if the provided state UUID exists in the database to prevent
        CSRF attacks. Raises an exception if the state is not found.

        Args:
            state_uuid: State UUID to verify

        Raises:
            StateNotFoundError: If the state is not found in the database
        """
        state_model = self.__models.state_class

        with self.__session_factory() as sess:
            if not state_model.exists(sess, state_uuid):
                raise StateNotFoundError()

    def __post_to_token_url(self, code: str) -> dict[str, str]:
        """
        Exchange authorization code for access token at OIDC token endpoint.

        Makes a POST request to the OIDC provider's token endpoint using HTTP Basic
        authentication with the authorization code to get an access token.

        Args:
            code: Authorization code from OIDC provider

        Returns:
            dict: Token response from OIDC provider

        Raises:
            requests.HTTPError: If the token request fails
        """
        client = HttpClient()

        session = client.get_session()
        r = session.post(
            self.__config.token_url,
            data=self.__token_post_data(code),
            auth=self.__basic_auth()
        )

        r.raise_for_status()
        return r.json()

    def __basic_auth(self) -> HTTPBasicAuth:
        """
        Create HTTP Basic Authentication object for OIDC client credentials.

        Returns:
            HTTPBasicAuth: Authentication object with client ID and secret
        """
        return HTTPBasicAuth(self.__config.client_id, self.__config.client_secret)

    def __token_post_data(self, code: str) -> dict[str, str]:
        """
        Create the POST data for token exchange request.

        Builds the form data required for the OIDC authorization code grant.

        Args:
            code: Authorization code to exchange

        Returns:
            dict: Form data for token request
        """
        return {
            'grant_type': 'authorization_code',
            'code': code,
            'redirect_uri': self.__config.redirect_uri,
        }

    def __delete_old_states(self) -> None:
        """
        Delete expired OIDC state entries from the database.

        Removes state entries that are older than the configured state_delete_delta
        to prevent accumulation of old state records.
        """
        before = datetime.now() - self.__state_delta
        state_model = self.__models.state_class

        with self.__session_factory() as sess:
            state_model.delete_old(sess, before)

    def __cleanup_before_login(self) -> None:
        """
        Perform cleanup operations before initiating a new login.

        Deletes old state entries and expired tokens to maintain database hygiene.
        """
        self.__delete_old_states()
        self.__delete_expired_tokens()

    def __delete_expired_tokens(self) -> None:
        """
        Remove expired access tokens from the database.

        Calls the token model's delete_expired method to clean up tokens
        that have passed their expiration time.
        """
        token_model = self.__models.token_class

        with self.__session_factory() as sess:
            token_model.delete_expired(sess)

    def __create_new_state(self) -> str:
        """
        Generate and store a new OIDC state parameter.

        Creates a new unique state UUID for CSRF protection during the OIDC flow
        and stores it in the database.

        Returns:
            str: The generated state UUID
        """
        state_model = self.__models.state_class

        with self.__session_factory() as sess:
            return state_model.add(sess)

    def __format_login_url(self, state_uuid: str) -> dict[str, str]:
        """
        Format the complete OIDC authorization URL with all required parameters.

        Constructs the authorization URL with the state parameter and other
        OIDC parameters like client_id, response_type, redirect_uri, and scope.

        Args:
            state_uuid: State parameter for CSRF protection

        Returns:
            dict: Dictionary containing 'loginUrl' with the complete authorization URL
        """
        encoded = self.__encode_params(state_uuid)
        login_url = f'{self.__config.auth_url}?{encoded}'

        return {'loginUrl': login_url}

    def __encode_params(self, state_uuid: str) -> str:
        """
        URL encode the OIDC authorization parameters.

        Creates and encodes the query parameters needed for the OIDC authorization request.

        Args:
            state_uuid: State parameter value

        Returns:
            str: URL-encoded query string with all OIDC parameters
        """
        params = {
            'client_id': self.__config.client_id,
            'response_type': 'code',
            'state': state_uuid,
            'redirect_uri': self.__config.redirect_uri,
            'scope': 'openid profile email',
        }

        return urlencode(params)

    def __get_membership(self, user: Any):
        pass


class DbAuthBlueprint(AuthBlueprint):
    """
    Flask Blueprint for database-backed authentication with OIDC.

    Extends the base `AuthBlueprint` to include database model classes
    for authentication. This blueprint provides all the standard auth
    endpoints while maintaining access to the underlying SQLAlchemy
    model classes used for user, token, and state management.

    The blueprint exposes the auth models through a `models` property,
    allowing other parts of the application to interact directly with
    the authentication database tables when needed.
    """

    def __init__(
        self, auth_manager: AuthManager, url_prefix: str, models: ModelTuple
    ) -> None:
        """
        Initialize the database authentication blueprint.

        Args:
            auth_manager: Authentication manager instance for handling auth operations
            url_prefix: URL prefix for all authentication endpoints
            models: Tuple containing the SQLAlchemy model classes for auth tables
        """
        super().__init__(auth_manager, url_prefix)

        self.__models = models

    @property
    def models(self) -> ModelTuple:
        """
        Get the authentication database model classes.

        Returns a `ModelTuple` containing the created `Model` classes for auth.
        These models can be used for direct database operations or for creating
        additional database relationships.

        Returns:
            ModelTuple: Container with the User, Token, State, and Role model classes
        """
        return self.__models


def __db_auth_manager(
    model_base: ModelClass,
    db_uri: str,
    oidc_config: OidcConfig,
    user_mixin_class: ModelClass,
    user_model_name: str,
    token_mixin_class: ModelClass,
    token_is_pk: bool,
    role_mixin_class: ModelClass,
    oidc_id_column_name: str,
    state_delete_delta: timedelta,
    token_expiry_delta: timedelta,
    oidc_id_target: str,
    oidc_ext_mapping: dict[str, str],
    prefix_with_name: bool,
    authorisation_manager: Optional[AuthorisationManager] = None
) -> tuple[DbAuthManager, ModelTuple]:
    """
    Create a database authentication manager with all required components.

    This internal function sets up the complete database authentication system
    including session factory, model creation, and auth manager initialization.

    Args:
        model_base: SQLAlchemy declarative base class for models
        db_uri: Database connection URI
        oidc_config: OIDC configuration object
        user_mixin_class: Mixin class to extend the User model
        user_model_name: Name for the User model class
        token_mixin_class: Mixin class to extend the Token model
        token_is_pk: Whether token should be used as primary key
        role_mixin_class: Mixin class to extend the Role model
        oidc_id_column_name: Column name for OIDC identifier
        state_delete_delta: Time after which states are cleaned up
        token_expiry_delta: Time after which tokens expire
        oidc_id_target: OIDC field to use as user identifier
        oidc_ext_mapping: Mapping of OIDC fields to local attributes
        prefix_with_name: Whether to prefix model names

    Returns:
        tuple: A tuple containing the DbAuthManager instance and ModelTuple
    """
    session_factory = create_session_factory(db_uri)

    model_tuple = create_models(
        model_base,
        user_model_name,
        oidc_id_column_name,
        user_mixin_class,
        token_mixin_class,
        token_is_pk,
        role_mixin_class,
        token_expiry_delta,
        prefix_with_name,
    )

    auth_manager = DbAuthManager(
        oidc_config,
        session_factory,
        model_tuple,
        state_delete_delta,
        oidc_id_target,
        oidc_ext_mapping,
        authorisation_manager,
    )

    return auth_manager, model_tuple


def db_auth_blueprint(
    model_base: ModelClass,
    db_uri: str,
    authorisation_manager: Optional[AuthorisationManager] = None,
    url_prefix: str = '/auth',
    prefix_with_name: bool = False,
    oidc_config_factory: Callable[[], OidcConfig] = env_oidc_config,
    user_mixin_class: ModelClass = object,
    user_model_name: str = 'user',
    token_mixin_class: ModelClass = object,
    token_is_pk: bool = False,
    role_mixin_class: ModelClass = object,
    oidc_id_column_name: str = 'oidc_id',
    state_delete_delta: timedelta = timedelta(hours=1),
    token_expiry_delta: timedelta = timedelta(days=7),
    oidc_id_target: str = 'email',
    oidc_ext_mapping: dict[str, str] = {},
) -> DbAuthBlueprint:
    """
    Create a Flask Blueprint for database-backed authentication with OIDC.

    Creates a complete authentication system using a database for persistence
    and OIDC for user identity verification. The function sets up all necessary
    database models, creates the authentication manager, and returns a Flask
    Blueprint ready to be registered with a Flask application.

    The blueprint provides standard authentication endpoints for login, callback,
    token management, and user profile access. It automatically handles token
    expiration, state cleanup, and user profile synchronization with OIDC.

    Args:
        model_base: SQLAlchemy declarative base class for creating models
        db_uri: Database connection URI (e.g., 'postgresql://user:pass@host/db')
        url_prefix: URL prefix for all authentication endpoints (default: '/auth')
        prefix_with_name: Whether to prefix model table names with the user model name
        oidc_config_factory: Factory function returning OIDC configuration
        user_mixin_class: Optional mixin class to extend the User model with additional
                         columns. All additional columns must be nullable or have defaults.
        user_model_name: Name for the generated User model class (default: 'user')
        token_mixin_class: Optional mixin class to extend the Token model
        token_is_pk: Whether the token value should be used as the primary key
        role_mixin_class: Optional mixin class to extend the Role model
        oidc_id_column_name: Column name for storing OIDC user identifier
        state_delete_delta: Time after which OIDC state entries are cleaned up
                           (default: 1 hour)
        token_expiry_delta: Time after which access tokens expire (default: 7 days)
        oidc_id_target: Field name in OIDC user info response to use as user ID
                       (default: 'email')
        oidc_ext_mapping: Dictionary mapping OIDC response field names to local
                         user attribute names for additional user information

    Returns:
        DbAuthBlueprint: Flask Blueprint instance containing authentication endpoints
                        and providing access to the generated database model classes

    Note:
        The returned blueprint must be registered with a Flask application using
        `app.register_blueprint(blueprint)` before the authentication endpoints
        will be available.

        Database tables will be created automatically when the models are first
        used. Ensure your database exists and is accessible via the provided URI.
    """
    auth_manager, models = __db_auth_manager(
        model_base,
        db_uri,
        oidc_config_factory(),
        user_mixin_class,
        user_model_name,
        token_mixin_class,
        token_is_pk,
        role_mixin_class,
        oidc_id_column_name,
        state_delete_delta,
        token_expiry_delta,
        oidc_id_target,
        oidc_ext_mapping,
        prefix_with_name,
        authorisation_manager=authorisation_manager,
    )

    auth_bp = DbAuthBlueprint(auth_manager, url_prefix, models)

    return auth_bp



# SPDX-FileCopyrightText: 2024 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

from abc import ABC, abstractmethod
from datetime import datetime, timedelta
from typing import Any, NamedTuple, Optional
from uuid import uuid4

from sqlalchemy import (
    ForeignKey,
    String,
    delete,
    select
)
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.orm import (
    Mapped,
    Session,
    joinedload,
    mapped_column,
    relationship
)


ModelClass = type[Any]


class AuthUser(ABC):
    """
    Adds authentication ability to a separate `User` class.
    """

    id: int  # noqa A003

    _tokens: list[AuthToken]
    _roles: list[AuthRole]

    @classmethod
    @abstractmethod
    def get_or_create(
        cls,
        sess: Session,
        oidc_id: str,
        **oidc_ext
    ) -> AuthUser:
        """
        Adds a new `User` row to the DB, with the
        given `oidc_id`, if it doesn't already exist.

        Returns the found row if it does.
        """

    @property
    @abstractmethod
    def role_names(self) -> list[str]:
        """
        A `list[str]` of roles assigned to this user.
        """


class AuthState(ABC):
    """
    Stores state between requests within the same
    OIDC handshake.
    """

    @classmethod
    @abstractmethod
    def add(cls, sess: Session) -> str:
        """
        Adds a `State` row to the DB, with a generated
        UUID. Returns the UUID `str`.
        """

    @classmethod
    @abstractmethod
    def exists(cls, sess: Session, uuid: str) -> str:
        """
        Returns `True` if a state with the given `uuid`
        exists.
        """

    @classmethod
    @abstractmethod
    def delete_old(cls, sess: Session, before: datetime) -> None:
        """
        Deletes all `State` rows older than the given
        `before`.
        """


class AuthToken(ABC):
    """Stores a token against a user."""

    user_id: str
    user: Optional[AuthUser]

    @classmethod
    @abstractmethod
    def get(
        cls,
        sess: Session,
        token: str
    ) -> Optional[AuthToken]:
        """
        Gets the `Token` instance with given value. Returns
        `None` if none is found.
        """

    @classmethod
    @abstractmethod
    def get_or_create(
        cls,
        sess: Session,
        token: str,
        user_id: int
    ) -> dict[str, str]:
        """
        Registers the token value against the `User` row
        of given `user_id`. Returns extra details about the
        token, as a `dict[str, str]`.
        """

    @classmethod
    @abstractmethod
    def delete(cls, sess: Session, token: str) -> None:
        """Deletes the row for the given `token`."""

    @classmethod
    @abstractmethod
    def delete_expired(cls, sess: Session) -> None:
        """Deletes expired tokens."""


class AuthRole:
    """A role, any number of which can be assigned to a user."""

    name: str


class AuthRoleBinding:
    """An assignment of a role to a user."""

    user_id: str
    role_id: str


class ModelTuple(NamedTuple):
    """
    Contains the `ModelClass` variables required
    for OIDC in a database.
    """

    state_class: type[AuthState]
    user_class: type[AuthUser]
    token_class: type[AuthToken]
    role_class: type[AuthRole]
    role_binding_class: type[AuthRoleBinding]


def create_models(
    model_base: ModelClass,
    user_table_name: str,
    oidc_id_column_name: str,
    user_mixin_class: ModelClass,
    token_mixin_class: ModelClass,
    token_is_pk: bool,
    role_mixin_class: ModelClass,
    token_expiry_delta: timedelta,
    prefix_with_name: bool
) -> ModelTuple:
    """
    Creates the OIDC db models, given a suitable base (and
    the name of the `User` model.)

    Returns a `NamedTuple`.
    """

    class State(AuthState, model_base):

        __tablename__ = 'oidc_state'

        id: Mapped[str] = mapped_column(  # noqa A003
            primary_key=True,
            name='state_id' if prefix_with_name else None
        )

        created_at: Mapped[datetime] = mapped_column(
            nullable=False,
            default=lambda: datetime.now()
        )

        @classmethod
        def add(cls, sess: Session) -> str:
            state_uuid = uuid4().hex

            sess.add(
                cls(id=state_uuid)
            )
            sess.commit()

            return state_uuid

        @classmethod
        def exists(
            cls,
            sess: Session,
            uuid: str
        ) -> str:

            stmt = select(cls).where(
                cls.id == uuid
            )
            rows = sess.execute(stmt)

            return len(list(rows)) > 0

        @classmethod
        def delete_old(
            cls,
            sess: Session,
            before: datetime
        ) -> None:

            stmt = delete(cls).where(
                cls.created_at < before
            )
            sess.execute(stmt)
            sess.commit()

    oidc_id_column: Mapped[str] = mapped_column(
        type_=String(),
        unique=True,
        nullable=False
    )

    oidc_id_mixin_class = type(
        'OidcIdMixin',
        (object,),
        {
            oidc_id_column_name: oidc_id_column,
            '__abstract__': True
        }
    )

    class User(
        AuthUser,
        model_base,
        oidc_id_mixin_class,
        user_mixin_class,
    ):

        __tablename__ = user_table_name

        id: Mapped[int] = mapped_column(  # noqa A003
            primary_key=True,
            autoincrement=True,
            name='user_id' if prefix_with_name else None
        )

        _tokens: Mapped[list[Token]] = relationship(
            back_populates='user'
        )
        _role_bindings: Mapped[list[RoleBinding]] = relationship(
            back_populates='user'
        )

        @classmethod
        def __get_oidc_id_column_name(cls) -> str:
            """
            Reliably gets the ID column of this `User` against
            the OIDC provider - irrespective of its name.
            """
            return getattr(cls, oidc_id_column_name).name

        @classmethod
        def __one_or_none(
            cls,
            sess: Session,
            oidc_id: str
        ) -> Optional[User]:

            name = cls.__get_oidc_id_column_name()

            return sess.query(
                cls
            ).filter_by(
                **{name: oidc_id}
            ).one_or_none()

        @classmethod
        def __add(
            cls,
            sess: Session,
            oidc_id: str,
            **oidc_ext
        ) -> User:

            name = cls.__get_oidc_id_column_name()

            new_user = cls(
                **oidc_ext,
                **{name: oidc_id},
            )
            sess.add(new_user)
            sess.commit()

            return new_user

        @classmethod
        def get_or_create(
            cls,
            sess: Session,
            oidc_id: str,
            **oidc_ext
        ) -> User:

            user = cls.__one_or_none(
                sess,
                oidc_id
            )

            if user is not None:
                return user

            return cls.__add(
                sess,
                oidc_id,
                **oidc_ext
            )

        @property
        def role_names(self) -> list[str]:
            names = [
                b.role.name for b in self._role_bindings
            ]
            return sorted(names)

    class _TokenPKMixin:
        """
        Has an integer as a primary key.

        This is all that is needed in most apps,
        and is the default.
        """

        __abstract__ = True

        @declared_attr
        def id(self) -> Mapped[int]:  # noqa A003
            return mapped_column(
                primary_key=True,
                autoincrement=True
            )

    class _EmptyTokenMixin(object):
        """Is deliberately empty"""

    token_pk_mixin_class = (
        _EmptyTokenMixin if token_is_pk else _TokenPKMixin
    )

    class Token(
        AuthToken,
        model_base,
        token_pk_mixin_class,
        token_mixin_class
    ):

        __tablename__ = 'token'

        token: Mapped[str] = mapped_column(
            nullable=False,
            unique=True,
            primary_key=token_is_pk
        )

        created_at: Mapped[datetime] = mapped_column(
            nullable=False,
            default=lambda: datetime.now()
        )
        expires_at: Mapped[datetime] = mapped_column(
            nullable=False,
            default=lambda: datetime.now() + token_expiry_delta
        )

        user_id: Mapped[int] = mapped_column(
            ForeignKey(User.id)
        )

        user = relationship(
            'User',
            back_populates='_tokens',
            foreign_keys=[user_id]
        )

        @classmethod
        def get(
            cls,
            sess: Session,
            token: str
        ) -> Optional[Token]:

            return (
                sess.query(cls)
                .filter_by(token=token)
                .options(
                    # Join through to the Role table when fetching a Token so
                    # that SQLAlchemy only fires one SELECT query instead of
                    # four to get `token.user._role_bindings.role.name`
                    joinedload(cls.user)
                    .joinedload(User._role_bindings)
                    .joinedload(RoleBinding.role)
                )
                .one_or_none()
            )

        @classmethod
        def get_or_create(
            cls,
            sess: Session,
            token: str,
            user_id: int
        ) -> dict[str, str]:

            token_row = cls.get(
                sess,
                token
            )

            if token_row is None:

                token_row = cls(
                    token=token,
                    user_id=user_id
                )
                sess.add(token_row)
                sess.commit()

            return token_row.__to_dict()

        @classmethod
        def delete(
            cls,
            sess: Session,
            token: str
        ) -> None:

            sess.query(cls).filter_by(token=token).delete()
            sess.commit()

        @classmethod
        def delete_expired(cls, sess: Session) -> None:
            stmt = delete(cls).where(
                cls.expires_at < datetime.now()
            )
            sess.execute(stmt)
            sess.commit()

        def __to_dict(self) -> dict[str, str]:
            return {
                'token_created_at': self.__str_datetime(
                    self.created_at
                ),
                'token_expires_at': self.__str_datetime(
                    self.expires_at
                ),
            }

        def __str_datetime(self, val: datetime) -> str:
            return val.strftime('%Y-%m-%dT%H:%M:%S.%f')

    class Role(
        AuthRole,
        model_base,
        role_mixin_class
    ):

        __tablename__ = 'role'

        id: Mapped[int] = mapped_column(  # noqa A003
            primary_key=True,
            autoincrement=True,
            name='role_id' if prefix_with_name else None
        )

        name: Mapped[str] = mapped_column(
            unique=True
        )

        _role_bindings: Mapped[list[RoleBinding]] = relationship(
            back_populates='role'
        )

    class RoleBinding(AuthRoleBinding, model_base):

        __tablename__ = 'role_binding'

        id: Mapped[int] = mapped_column(  # noqa A003
            primary_key=True,
            autoincrement=True
        )

        user_id: Mapped[int] = mapped_column(
            ForeignKey(User.id)
        )
        role_id: Mapped[int] = mapped_column(
            ForeignKey(Role.id)
        )

        user = relationship(
            'User',
            back_populates='_role_bindings',
            foreign_keys=[user_id]
        )
        role = relationship(
            'Role',
            back_populates='_role_bindings',
            foreign_keys=[role_id]
        )

    return ModelTuple(
        state_class=State,
        token_class=Token,
        user_class=User,
        role_class=Role,
        role_binding_class=RoleBinding
    )



# SPDX-FileCopyrightText: 2021 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

import logging
from abc import ABCMeta, abstractmethod

from lab_share_lib.exceptions import TransientRabbitError
from lab_share_lib.processing.rabbit_message import RabbitMessage
from lab_share_lib.processing.rabbit_message_processor import ENCODERS
from lab_share_lib.rabbit.schema_registry import SchemaRegistry

import tol.sciops.configuration as config

LOGGER = logging.getLogger(__name__)


class MessageProcessor(object, metaclass=ABCMeta):
    """ Base class for response message processors """

    @property
    def registry(self):
        return self.__registry

    def __init__(self):
        self.__registry = SchemaRegistry(config.REDPANDA_URL, config.REDPANDA_API_KEY)

    def build_avro_encoder(self, encoder_type, subject):
        if encoder_type not in ENCODERS.keys():
            raise Exception(f'Encoder type {encoder_type} not recognised')

        return ENCODERS[encoder_type](self.__registry, subject)

    @abstractmethod
    def process_message(self, headers, body):
        pass

    def _unpack(self, headers, body) -> RabbitMessage:
        """ Helper method to decode a received feedback message """
        message = RabbitMessage(headers, body)
        message.decode(self.build_avro_encoder(message.encoder_type, message.subject))
        return message


class FeedbackProcessor(MessageProcessor):
    """ Handle incoming feedback messages """

    def process_message(self, headers, body):
        """
        This method should return true to ack the message and False to move it to dead letter
        """
        try:
            message = self._unpack(headers, body)
        except TransientRabbitError as ex:
            # Cause the consumer to restart and try this message again.
            # Ideally we will delay the consumer.
            LOGGER.error(f'Transient error while processing message: {ex.message}')
            raise
        except Exception as ex:
            LOGGER.error(
                f'Unrecoverable error while decoding RabbitMQ message: {type(ex)} {str(ex)}')
            return False  # Send the message to dead letters.

        if not message.contains_single_message:
            LOGGER.error('RabbitMQ message received containing multiple AVRO encoded messages.')
            return False  # Send the message to dead letters.

        LOGGER.info(f'Received queue feedback message: {str(message.message)}')
        return True


class NoOpProcessor(MessageProcessor):
    """ No operation dummy feedback processor for testing """

    def process_message(self, headers, body):
        return True



# SPDX-FileCopyrightText: 2025 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

import re
from dataclasses import dataclass

from tol.core import DataSource
from tol.core.data_object import DataObject
from tol.core.validate import Validator
from tol.sources.ena import ena


class EnaChecklistValidator(Validator):
    ENA_TO_INTERNAL_FIELD_MAP = {
        'geographic location (longitude)': 'DECIMAL_LONGITUDE',
        'geographic location (latitude)': 'DECIMAL_LATITUDE',
    }

    """
    validates the ENA_CHECKLIST for each samples
    """

    @dataclass(slots=True, frozen=True, kw_only=True)
    class Config:
        ena_checklist_id: str

    __slots__ = ['__config']
    __config: Config

    def __init__(self, config: Config, datasource: DataSource = ena(), **kwargs) -> None:
        super().__init__()
        self.__config = config
        self._datasource = datasource
        self.__ena_checklist = datasource.get_one(
            'checklist',
            self.__config.ena_checklist_id
        ).checklist

    def _validate_data_object(self, obj: DataObject) -> None:
        for key, validation in self.__ena_checklist.items():
            field_name = key
            if 'field' in validation:
                field_name = validation['field']
            internal_field = self.ENA_TO_INTERNAL_FIELD_MAP.get(field_name, field_name)

            if 'mandatory' in validation and key not in obj.attributes:
                self.add_error(object_id=obj.id, detail='Must be given', field=[internal_field])
                continue
            if 'mandatory' in validation and obj.attributes[key] is None:
                self.add_error(object_id=obj.id, detail='Must be given', field=[internal_field])
                continue
            if 'mandatory' in validation and obj.attributes.get(key) == '':
                self.add_error(
                    object_id=obj.id,
                    detail='Must not be empty', field=[internal_field]
                )

            if 'restricted text' in validation and key in obj.attributes:
                for condition in validation:
                    if isinstance(condition, str) and '(' in condition:
                        regex = condition
                compiled_re = re.compile(regex)
                if not compiled_re.search(obj.attributes.get(key)):
                    if internal_field in ('DECIMAL_LONGITUDE', 'DECIMAL_LATITUDE'):
                        self.add_error(
                            object_id=obj.id,
                            detail=(
                                f'Value for {internal_field} must be a valid decimal in the'
                                f' format required by ENA (e.g. 12.3456). '
                                f'See ENA checklist for accepted format.',
                            ),
                            field=[internal_field]
                        )
                    else:
                        self.add_error(
                            object_id=obj.id,
                            detail=(
                                f'Value for {internal_field} must match the required pattern.'
                            ),
                            field=[internal_field]
                        )

            if 'text choice' in validation and key in obj.attributes:
                for condition in validation:
                    if isinstance(condition, list):
                        allowed_values = condition
                if obj.attributes.get(key).lower() not in \
                        [x.lower() for x in allowed_values]:
                    self.add_error(
                        object_id=obj.id,
                        detail='Must be in allowed values', field=[internal_field]
                    )



# SPDX-FileCopyrightText: 2022 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT


import xml.etree.ElementTree as ElementTree
from functools import cache
from typing import Callable, Dict, Iterable, Optional, Tuple

from cachetools.func import ttl_cache

import requests
from requests.auth import HTTPBasicAuth
from requests.exceptions import HTTPError


from .client import EnaApiClient
from .converter import (
    EnaApiConverter
)
from .ena_methods import (
    assign_ena_ids,
    build_bundle_sample_xml,
    build_submission_xml,
    convert_checklist_xml_to_dict,
    convert_xml_to_list_of_sample_dict,
)
from .filter import (
    EnaFilter
)
from ..core import (
    DataObject,
    DataSource,
    DataSourceError,
    DataSourceFilter
)
from ..core.operator import (
    DetailGetter,
    ListGetter,
    PageGetter,
    Relational
)
from ..core.relationship import RelationshipConfig

ClientFactory = Callable[[], EnaApiClient]
FilterFactory = Callable[[], EnaFilter]
EnaConverterFactory = Callable[[], EnaApiConverter]


class EnaDataSource(
    DataSource,

    # the supported operators
    DetailGetter,
    ListGetter,
    PageGetter,
    Relational
):
    """
    A `Datasource` that connects to a remote ENA API.

    Developers should likely use `create_ena_data_source`
    instead of this directly
    """
    def __init__(
        self,
        client_factory: ClientFactory,
        ena_converter_factory: EnaConverterFactory,
        filter_factory: FilterFactory
    ) -> None:

        self.__client_factory = client_factory
        self.__ec_factory = ena_converter_factory
        self.__filter_factory = filter_factory
        super().__init__({})

    @ttl_cache(ttl=60)
    def get_fields(self, object_type) -> dict:
        return self.__client_factory().get_fields(object_type)

    def __get_filter_string(
        self,
        object_filters=Optional[DataSourceFilter]
    ) -> Optional[str]:
        if object_filters is None:
            return ''

        return self.__filter_factory().dumps(object_filters)

    @property
    @cache
    def attribute_types(self) -> dict[str, dict[str, str]]:
        att_types = {}

        for object_type in self.supported_types:
            fields = self.get_fields(object_type)
            att_types[object_type] = fields

        return att_types

    @property
    @cache
    def supported_types(self) -> list[str]:
        return [
            'assembly', 'read_run', 'sample', 'study', 'taxon', 'checklist',
            'submittable_taxon'
        ]

    def get_by_id(
        self,
        object_type: str,
        object_ids: Iterable[str],
        **kwargs,
    ) -> Iterable[Optional[DataObject]]:
        self.__validate_object_type(object_type)

        client = self.__client_factory()
        # For a submittable_taxon we need to make multiple calls
        if object_type == 'submittable_taxon':
            ena_response = []
            for object_id in object_ids:
                try:
                    response = client.get_detail(object_type, [object_id])
                    if response and isinstance(response, list):
                        ena_response.extend(response)
                except HTTPError as http_error:
                    if http_error.response.status_code != 400:
                        raise
        else:
            ena_response = client.get_detail(object_type, object_ids)
        # For a checklist we need to convert into a list of dicts
        if object_type == 'checklist':
            ena_response = convert_checklist_xml_to_dict(ena_response)
        ena_converter = self.__ec_factory()

        converted_objects, _ = ena_converter.convert_list(object_type, ena_response) \
            if ena_response is not None else ([], 0)
        yield from self.sort_by_id(converted_objects, object_ids)

    def get_list_page(
        self,
        object_type: str,
        page_number: int,
        page_size: Optional[int] = None,
        object_filters: Optional[DataSourceFilter] = None,
        sort_by: Optional[str] = None
    ) -> tuple[Iterable[DataObject], int]:

        size = page_size if page_size else self.get_page_size()
        filter_string = self.__get_filter_string(object_filters)

        self.__validate_object_type(object_type)
        client = self.__client_factory()
        ena_response = client.get_list(
            object_type,
            filter_string=filter_string
        )

        converted_objects, total = self.__ec_factory().convert_list(object_type, ena_response)
        # sort by id by default
        sorted_response = sorted(converted_objects, key=lambda x: x.id)
        list_slice_objects = sorted_response[((page_number - 1) * size):(page_number * size)]

        return list_slice_objects, total

    def get_list(
        self,
        object_type: str,
        object_filters: Optional[DataSourceFilter] = None,
        sort_by: Optional[str] = None
    ) -> Iterable[DataObject]:

        if sort_by is not None:
            self.__raise_sort_by_error()

        self.__validate_object_type(object_type)
        filter_string = self.__get_filter_string(object_filters)
        client = self.__client_factory()

        ena_response = client.get_list(
            object_type,
            filter_string=filter_string
        )

        converted_objects, total = self.__ec_factory().convert_list(object_type, ena_response) \
            if ena_response is not None else ([], 0)
        yield from converted_objects

    @property
    @cache
    def relationship_config(self) -> dict[str, RelationshipConfig]:
        return {
        }

    def get_to_one_relation(
        self,
        source: DataObject,
        relation_name: str,
    ) -> Optional[DataObject]:
        return None

    def get_to_many_relations(
        self,
        source: DataObject,
        relation_name: str,
    ) -> Iterable[DataObject]:
        return []

    def post_request(self, command: str, files) -> requests.Response:
        response = requests.post(self.uri + command,
                                 files=files,
                                 auth=HTTPBasicAuth(self.user, self.password))
        if (response.status_code != 200):
            raise DataSourceError(title='Cannot connect to ENA',
                                  detail=f"(status code '{str(response.status_code)}')'")

        return response

    def get_request(self, command: str, headers=None, params=None) -> requests.Response:
        response = requests.get(self.uri + command,
                                params=params, headers=headers,
                                auth=HTTPBasicAuth(self.user, self.password))

        if (response.status_code != 200):
            raise DataSourceError(title='Cannot connect to ENA',
                                  detail=f"(status code '{str(response.status_code)}')'")

        return response

    def get_biosample_data_biosampleid(self, biosample_id: str):
        output = self.get_request(f'/ena/browser/api/xml/{biosample_id}')

        samples = convert_xml_to_list_of_sample_dict(output.text)

        # Only returning one sample for biosample
        return samples[0]

    def generate_ena_ids_for_samples(self, manifest_id: str,
                                     samples: Dict[str, Dict]) -> Tuple[str, Dict[str, Dict]]:

        bundle_xml_file, sample_count = build_bundle_sample_xml(samples)

        with open(bundle_xml_file, 'r') as bxf:
            bundle_xml_file_contents = bxf.read()

            element = ElementTree.XML(bundle_xml_file_contents)
            ElementTree.indent(element)
            bundle_xml_file_contents = ElementTree.tostring(element, encoding='unicode')

        if sample_count == 0:
            raise DataSourceError(title='All samples have unknown taxonomy ID',
                                  detail='')

        submission_xml_file = build_submission_xml(manifest_id, self.contact_name,
                                                   self.contact_email)

        xml_files = [('SAMPLE', open(bundle_xml_file, 'rb')),
                     ('SUBMISSION', open(submission_xml_file, 'rb'))]

        response = self.post_request('/ena/submit/drop-box/submit/', xml_files)

        try:
            assigned_samples = assign_ena_ids(samples, response.text)

        except Exception as ex:
            raise DataSourceError(title='Error returned from ENA service',
                                  detail=ex)

        if not assigned_samples:
            errors = {}
            error_count = 0
            for error_node in ElementTree.fromstring(response.text).findall('./MESSAGES/ERROR'):
                if error_node is not None:
                    error_count += 1
                    errors[str(error_count)] = error_node.text

            return False, errors
        else:
            return True, assigned_samples

    def __validate_object_type(self, object_type: str):
        if object_type not in self.supported_types:
            raise DataSourceError(
                title='Unsupported object type',
                detail=f"Object type '{object_type}' is not supported by this data source."
            )



# SPDX-FileCopyrightText: 2023 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from typing import Dict, Optional

from .filter_utils import FilterUtils


class AggregationParameters:
    """
    Parses the parameters from a query string for a Aggregation POST
    endpoint.
    """

    def __init__(self, request_args: Dict[str, str]) -> None:
        self.__request_args = request_args

    @property
    def filter(self) -> Optional[str]:  # noqa A003
        """
        The optional filter JSON string.
        """
        filter_string = self.__request_args.get('filter')
        if filter_string is None:
            return None

        return FilterUtils.parse_to_datasource_filter('filter', filter_string)



# SPDX-FileCopyrightText: 2022 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from typing import Any, Dict, List, Optional, Type

from ..core import DataSource


class BaseRuntimeException(Exception):
    def __init__(
        self,
        errors: List[Dict[str, Any]],
        status_code: int = 500
    ) -> None:
        self.__errors = errors
        self.__status_code = status_code

    @property
    def errors(self) -> List[Dict[str, Any]]:
        """
        The list of errors, containing a title and detail each
        """
        return self.__errors

    @property
    def status_code(self) -> int:
        """
        The status code for the error response in Flask
        """
        return self.__status_code


class ObjectNotFoundByIdException(BaseRuntimeException):
    def __init__(self, object_type: str, id_: str) -> None:
        errors = [{
            'title': 'Object Not Found',
            'detail': f'No "{object_type}" object was found with id "{id_}".'
        }]
        super().__init__(errors, status_code=404)


class RecursiveRelationNotFoundException(BaseRuntimeException):
    def __init__(self) -> None:
        errors = [{
            'title': 'Relation Not Found',
            'detail': (
                'The specified to-one relation object was not found'
            )
        }]
        super().__init__(errors, status_code=404)


class UnsupportedOperationError(BaseRuntimeException):
    def __init__(self, object_type: str, operation: str) -> None:
        errors = [{
            'title': 'Unsupported Operation',
            'detail': f'Cannot perform "{operation}" on type "{object_type}".'
        }]
        super().__init__(errors, status_code=400)


class UninheritedOperationError(BaseRuntimeException):
    def __init__(
        self,
        data_source: DataSource,
        operator_class: Type,
        method: str
    ) -> None:

        self.__detail = (
            f'The DataSource {type(data_source).__name__} must '
            f'inherit from {operator_class.__name__} to implement '
            f'the {method} method.'
        )
        errors = [{
            'title': 'Misconfigured DataSource',
            'detail': self.__detail
        }]
        super().__init__(errors, status_code=500)

    def __str__(self) -> str:
        return self.__detail


class BadQueryArgError(BaseRuntimeException):
    def __init__(
        self,
        __key: str,
        __value: str,
        message: Optional[str] = None
    ) -> None:
        errors = [{
            'title': 'Bad Query-String Argument',
            'detail': self.__get_detail(__key, __value, message)
        }]
        super().__init__(errors, status_code=400)

    def __get_detail(
        self,
        __key: str,
        __value: str,
        message: Optional[str]
    ) -> str:
        detail = (
            f'The query-string argument with key "{__key}" '
            f'is invalid.'
        )
        if message is None:
            return detail
        return f'{detail}\n\n{message}'


class BadPostJsonError(BaseRuntimeException):
    def __init__(
        self,
        __key: str,
        message: Optional[str] = None
    ) -> None:
        errors = [{
            'title': 'Bad POST JSON',
            'detail': self.__get_detail(__key, message)
        }]
        super().__init__(errors, status_code=400)

    def __get_detail(
        self,
        __key: str,
        message: Optional[str]
    ) -> str:
        detail = (
            f'The POSTed JSON must have a key "{__key}"'
        )
        if message is None:
            return detail
        return f'{detail}\n\n{message}'


class UnauthenticatedError(BaseRuntimeException):
    """
    Raise on any error condition that warrants a
    401: Unauthorized HTTP status code.
    """

    def __init__(self, detail: str) -> None:
        errors = [{
            'title': 'Unauthorized',
            'detail': detail
        }]
        super().__init__(errors, status_code=401)



# SPDX-FileCopyrightText: 2025 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

from collections.abc import Iterable, Iterator

from . import DataSourceError, OperableDataSource
from .operator import Relational
from .relationship import RelationshipConfig


def requested_fields_to_tree(func):
    """
    Allows `requested_fields` keyword arguments to be supplied to methods if a
    `requested_tree` object has not been given.
    """

    def wrapper(self, object_type, *args, **kwargs):
        if 'requested_fields' in kwargs:
            if 'requested_tree' in kwargs:
                msg = 'Both requested_fields and requested_tree arguments given'
                raise TypeError(msg)
            flds = kwargs.pop('requested_fields')
            if flds and object_type:
                kwargs['requested_tree'] = ReqFieldsTree(
                    object_type, self, requested_fields=flds
                )
        return func(self, object_type, *args, **kwargs)
    return wrapper


class ReqFieldsTree:
    """
    Acts as a template for which related objects and attributes to fetch from
    the DataSource, and for serialzing them in the response.
    """

    def __init__(
        self,
        object_type: str,
        data_source: OperableDataSource,
        requested_fields: list[str] | None = None,
        include_all_to_ones: bool = True,
    ) -> None:
        self.__object_type: str = object_type
        self.__attributes: dict[str, bool] = {}
        self.__sub_trees: dict[str, ReqFieldsTree] = {}
        self.__rel_conf: RelationshipConfig | None = (
            data_source.relationship_config.get(object_type)
            if isinstance(data_source, Relational)
            else None
        )
        if requested_fields:
            self.add_requested_tree(data_source, requested_fields)
        elif include_all_to_ones:
            self.add_all_to_ones(data_source)

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.__dict__ == other.__dict__
        return False

    @property
    def object_type(self) -> str:
        return self.__object_type

    @property
    def attribute_names(self) -> list[str]:
        return list(self.__attributes)

    def add_attribute(self, name: str) -> None:
        """
        Add attribute `name`. Attributes are implemented as a dict rather than
        a set to preserve order.
        """
        self.__attributes[name] = True

    def has_attribute(self, name: str) -> bool:
        return name in self.__attributes

    def add_sub_tree(self, name: str, sub: ReqFieldsTree) -> None:
        self.__sub_trees[name] = sub

    def get_sub_tree(self, name: str) -> ReqFieldsTree | None:
        return self.__sub_trees.get(name)

    def sub_trees(self) -> Iterator[tuple[str, ReqFieldsTree]]:
        """
        Iterator over `name, sub_tree`.
        """
        yield from self.__sub_trees.items()

    @property
    def is_leaf(self) -> bool:
        """
        This tree is a leaf if it has no sub-trees.
        """
        return not self.__sub_trees

    @property
    def is_stub(self) -> bool:
        """
        This tree is a stub if it has no sub-trees and has one attribute which
        is "id".
        """
        if self.__sub_trees:
            return False
        return len(self.__attributes) == 1 and 'id' in self.__attributes

    @property
    def has_relationships(self) -> bool:
        return self.__rel_conf is not None

    def to_one_names(self) -> Iterable[str]:
        return x.keys() if (x := self.__rel_conf.to_one) else ()

    def to_many_names(self) -> Iterable[str]:
        return x.keys() if (x := self.__rel_conf.to_many) else ()

    def get_relationship(self, name: str) -> str:
        """
        Fetches the related object type from the `to_one` or `to_many` fields
        of the attached `RelationshipConfig`.
        """
        if rel_conf := self.__rel_conf:
            for attr in 'to_one', 'to_many':
                if (cfg := getattr(rel_conf, attr)) and (type_name := cfg.get(name)):
                    return type_name

    def get_attribute_type(self, data_source, name) -> type | None:
        return data_source.attribute_types[self.object_type].get(name)

    def add_all_to_ones(self, data_source: OperableDataSource) -> None:
        if not self.__rel_conf:
            return
        for name in self.to_one_names():
            sub_type = self.get_relationship(name)
            self.add_sub_tree(
                name,
                self.__class__(
                    sub_type,
                    data_source,
                    include_all_to_ones=False,
                ),
            )

    def add_requested_tree(
        self,
        data_source: OperableDataSource,
        requested_fields: list[str],
    ) -> None:
        err_title = 'Bad Requested Fields Path Element'
        for path_str in requested_fields:
            tree = self
            for name in path_str.split('.'):
                if name == '':
                    msg = f"Empty element in path path '{path_str}'"
                    raise DataSourceError(title=err_title, detail=msg)
                elif not tree:
                    msg = f"Element '{name}' appears after an attribute name in path '{path_str}'"
                    raise DataSourceError(title=err_title, detail=msg)
                elif sub_tree := tree.get_sub_tree(name):
                    # `name` is a relationship we already have so move pointer
                    # to the sub tree.
                    tree = sub_tree
                elif tree.has_attribute(name):
                    # Already have attribute `name`. Unset `tree` to trap path
                    # elements follwing an attribute.
                    tree = None
                elif sub_type := tree.get_relationship(name):
                    # New sub-tree from relationship name.
                    sub_tree = self.__class__(
                        sub_type,
                        data_source,
                        include_all_to_ones=False,
                    )
                    tree.add_sub_tree(name, sub_tree)
                    tree = sub_tree
                elif tree.get_attribute_type(data_source, name):
                    tree.add_attribute(name)
                    # Unset `tree` to trap path elements follwing an
                    # attribute.
                    tree = None
                else:
                    msg = (
                        f'{name!r} in path {path_str!r} is not a known relationship'
                        f' or attribute of {tree.object_type!r} objects'
                    )
                    raise DataSourceError(title=err_title, detail=msg)

    def __str__(self):
        return ','.join(self.__to_strings_iter())

    def to_paths(self) -> list[str]:
        return list(self.__to_strings_iter())

    def __to_strings_iter(self, *path: list[str]) -> Iterator[str]:
        for name in self.__attributes:
            yield '.'.join([*path, name])
        if self.is_leaf:
            # We only need to return the path to self if it hasn't already
            # appeared as a root path under an attribtue.
            if path and not self.__attributes:
                yield '.'.join(path)
        else:
            for name, tree in self.sub_trees():
                yield from tree.__to_strings_iter(*path, name)



# SPDX-FileCopyrightText: 2026 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

import typing
from functools import cache
from typing import Callable, Iterable, Optional

from .client import DummyClient
from .converter import (
    DummyConverter
)
from ..core import DataObject, DataSource, DataSourceError, DataSourceFilter
from ..core.operator import (
    DetailGetter,
    ListGetter
)

if typing.TYPE_CHECKING:
    from ..core.session import OperableSession

ClientFactory = Callable[[], DummyClient]
DummyConverterFactory = Callable[[], DummyConverter]


class DummyDataSource(
    DataSource,

    # the supported operators
    DetailGetter,
    ListGetter
):
    """
    A `DataSource` that outputs dummy data.

    Developers should likely use `create_dummy_datasource`
    instead of this directly.
    """

    def __init__(
        self,
        client_factory: ClientFactory,
        dummy_converter_factory: DummyConverterFactory
    ) -> None:

        self.__client_factory = client_factory
        self.__dummy_converter_factory = dummy_converter_factory
        super().__init__({})

    @property
    @cache
    def attribute_types(self) -> dict[str, dict[str, str]]:
        return {
            'record': {
                'big_string': 'str',
                'little_string': 'str',
                'bool': 'bool',
                'date': 'datetime',
                'int': 'int',
                'list': 'list[str]'
            },
            'category': {
                'name': 'str'
            }
        }

    @property
    @cache
    def supported_types(self) -> list[str]:
        return list(
            self.attribute_types.keys()
        )

    def get_by_id(
        self,
        object_type: str,
        object_ids: Iterable[str],
        **kwargs
    ) -> Iterable[Optional[DataObject]]:
        if object_type not in self.supported_types:
            raise DataSourceError(f'{object_type} is not supported')

        client = self.__client_factory()
        dummy_response = client.get_detail(object_type, object_ids)
        dummy_converter = self.__dummy_converter_factory()

        converted_objects, _ = dummy_converter.convert_list(dummy_response) \
            if dummy_response is not None else ([], 0)
        yield from self.sort_by_id(converted_objects, object_ids)

    def get_list(
        self,
        object_type: str,
        object_filters: Optional[DataSourceFilter] = None,
        session: Optional[OperableSession] = None
    ) -> Iterable[DataObject]:
        if object_filters:
            raise DataSourceError('Filtering is not supported')
        objects = self.__client_factory().get_list(
            object_type
        )
        converted_objects, _ = self.__dummy_converter_factory().convert_list(objects)
        return iter(converted_objects)



# SPDX-FileCopyrightText: 2025 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from dataclasses import dataclass

from tol.core import DataObject
from tol.core.validate import Validator


class MinOneValidValueValidator(Validator):
    """
    Validates that a stream of `DataObject` instances
    have at least one valid value present in a list of specified keys.
    """
    @dataclass(slots=True, frozen=True, kw_only=True)
    class Config:
        non_valid_values: list[str]
        keys: list[str]

    __slots__ = ['__config']
    __config: Config

    def __init__(
        self,
        config: Config,
        **kwargs
    ) -> None:

        super().__init__()
        self.__config = config

    def _validate_data_object(
        self,
        obj: DataObject
    ) -> None:

        found_valid_value = False

        for key in self.__config.keys:
            value = obj.attributes[key]

            if value is not None and value not in self.__config.non_valid_values:
                found_valid_value = True
                break

        if not found_valid_value:
            self.add_error(
                object_id=obj.id,
                detail=(
                    f'At least one of: {self.__config.keys} '
                    'must not be: ' + ', '.join(self.__config.non_valid_values)
                    + ' or empty.'
                ),
                field=', '.join(self.__config.keys),
            )



# SPDX-FileCopyrightText: 2023 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from typing import Iterable

from ...core import (
    DataObject,
    DataObjectToDataObjectOrUpdateConverter
)


class BenchlingTissuePrepToElasticTissuePrepConverter(
        DataObjectToDataObjectOrUpdateConverter):
    def convert(self, data_object: DataObject) -> Iterable[DataObject]:
        ret = self._data_object_factory(
            'tissue_prep',
            data_object.id,
            attributes={
                **{k: v
                   for k, v in data_object.attributes.items()
                   if k not in ['sts_id',
                                'taxon_id',
                                'programme_id']}
            },
            to_one={
                'sample': self._data_object_factory(
                    'sample',
                    data_object.sts_id
                ) if data_object.sts_id is not None else None,
                'species': self._data_object_factory(
                    'species',
                    data_object.taxon_id
                ) if data_object.taxon_id is not None else None,
                'tolid': self._data_object_factory(
                    'tolid',
                    data_object.programme_id
                ) if data_object.programme_id is not None else None,
            }
        )
        return iter([ret])



# SPDX-FileCopyrightText: 2023 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from collections.abc import Mapping
from typing import Callable, Iterator, Optional

from .api_datasource import ApiDataSource, DOConverterFactory, JsonConverterFactory
from .client import JsonApiClient
from .converter import DataObjectConverter, JsonApiConverter
from .filter import DefaultApiFilter
from .parser import DefaultParser
from ..core import DataSource, ReqFieldsTree


class _ApiDSDict(Mapping):
    def __init__(self, api_ds: ApiDataSource) -> None:
        self.__ds = api_ds

    def __getitem__(self, __k: str) -> ApiDataSource:
        if __k not in self.__ds.supported_types:
            raise KeyError()
        return self.__ds

    def __iter__(self) -> Iterator[str]:
        return iter(self.__ds.supported_types)

    def __len__(self) -> int:
        return len(self.__ds.supported_types)


class _ConverterFactory:
    """
    Manges the instantation of:

    - `DataObjectConverter`
    - `JsonApiConverter`
    """

    def __init__(self, prefix: str) -> None:
        self.__prefix = prefix
        self.__data_source: Optional[DataSource] = None

    @property
    def data_source(self) -> Optional[DataSource]:
        return self.__data_source

    @data_source.setter
    def data_source(self, ds: DataSource) -> None:
        self.__data_source = ds

    def do_converter_factory(self) -> DOConverterFactory:
        """
        Returns an instantiated `DataObjectConverter`.
        """

        return DataObjectConverter(self.__data_source, prefix=self.__prefix)

    def json_converter_factory(
        self,
        object_type: str | None = None,
        requested_tree: ReqFieldsTree | None = None,
    ) -> JsonConverterFactory:
        """
        Returns an instantiated `JsonApiConverter`.
        """

        # If we're going to parse a JSON:API data object (i.e. we have been
        # passed an `object_type`), we need to add a ReqFieldsTree so that
        # DataSources with default loading of to-one objects correctly
        # process the to-ones from the `included` array.
        if object_type and not requested_tree:
            requested_tree = ReqFieldsTree(object_type, self.__data_source)

        parser = DefaultParser(self.__ds_dict, requested_tree)
        return JsonApiConverter(parser)

    @property
    def __ds_dict(self) -> dict[str, DataSource]:
        return _ApiDSDict(self.data_source)


def _get_client_factory(
    api_url: str,
    token: Optional[str],
    data_prefix: str,
    retries: int,
    status_forcelist: list[int],
    merge_collections: bool | None,
) -> Callable[[], JsonApiClient]:
    """
    A resonable default for creating
    a `JsonApiClient` instance
    """

    return lambda: JsonApiClient(
        api_url,
        token=token,
        data_prefix=data_prefix,
        retries=retries,
        status_forcelist=status_forcelist,
        merge_collections=merge_collections,
    )


def _filter_factory() -> DefaultApiFilter:
    return DefaultApiFilter()


def create_api_datasource(
    api_url: str,
    token: Optional[str] = None,
    data_prefix: str = '/data',
    retries: int = 5,
    status_forcelist: Optional[list[int]] = None,
    merge_collections: bool | None = None,
) -> ApiDataSource:
    """
    Instantiates `ApiDataSource` using the given:

    - `api_url`
    - `token` (optional)
    """

    client_factory = _get_client_factory(
        api_url,
        token=token,
        data_prefix=data_prefix,
        retries=retries,
        status_forcelist=status_forcelist,
        merge_collections=merge_collections,
    )
    manager = _ConverterFactory(data_prefix)

    api_ds = ApiDataSource(
        client_factory,
        manager.json_converter_factory,
        manager.do_converter_factory,
        _filter_factory,
    )

    manager.data_source = api_ds

    return api_ds



# SPDX-FileCopyrightText: 2025 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

import os

from .defaults import Defaults
from ..core import (
    core_data_object
)
from ..s3 import (
    S3DataSource,
    create_s3_datasource
)


def bioscan_image(**kwargs) -> S3DataSource:
    s3ds = create_s3_datasource(
        bucket_name=os.getenv('BIOSCAN_IMAGE_BUCKET_NAME',
                              Defaults.BIOSCAN_IMAGE_BUCKET_NAME),
        prefix=os.getenv('BIOSCAN_IMAGE_PREFIX', Defaults.BIOSCAN_IMAGE_PREFIX),
    )
    core_data_object(s3ds)
    return s3ds



# SPDX-FileCopyrightText: 2022 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from .sanitise import sanitise_value


def flatten_entity(entity, level=None):
    flattened_entity = {}
    prefix = f'{level}_' if level is not None else ''
    if isinstance(entity, dict):
        for field, value in entity.items():
            extra_fields = flatten_entity(value, prefix + field)
            flattened_entity = {**flattened_entity,
                                **extra_fields}
    elif isinstance(entity, list):
        for index, subentity in enumerate(entity):
            new_name = prefix + str(index)
            extra_fields = flatten_entity(subentity, new_name)
            flattened_entity = {**flattened_entity,
                                **extra_fields}
    else:
        return {level: entity}
    return flattened_entity


def convert_sts_entity_to_eln_entity_fields(sts_entity, mapping):
    eln_entity = {}
    flattened_entity = flatten_entity(sts_entity)
    for sts_name, eln_name in mapping['field_mappings'].items():
        if sts_name in flattened_entity:
            eln_entity[eln_name] = {'value': sanitise_value(flattened_entity[sts_name], None)}
    return eln_entity



# SPDX-FileCopyrightText: 2023 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from typing import Iterable

from ...core import (
    DataObject,
    DataObjectToDataObjectOrUpdateConverter
)


class TreeofsexUploadToTreeofsexAttributeConverter(
        DataObjectToDataObjectOrUpdateConverter):

    def __init__(self, data_object_factory, user=None):
        super().__init__(data_object_factory)
        self.__sources_seen = set()
        self.__species_seen = set()
        self.__user = user

    def convert_iterable(self, inputs):
        return super().convert_iterable(inputs)

    def convert(self, data_object: DataObject) -> Iterable[DataObject]:
        # Have we seen the source before?
        if data_object.source not in self.__sources_seen:
            self.__sources_seen.add(data_object.source)
            yield self._data_object_factory(
                'source',
                data_object.source
            )

        # Have we seen the species before?
        if data_object.taxon_id not in self.__species_seen:
            self.__species_seen.add(data_object.taxon_id)
            yield self._data_object_factory(
                'species',
                data_object.taxon_id
            )

        yield self._data_object_factory(
            'attribute',
            None,
            attributes={
                'value': data_object.attribute_value,
                'state': data_object.attribute_state
            },
            to_one={
                'source': self._data_object_factory(
                    'source',
                    data_object.source
                ),
                'species': self._data_object_factory(
                    'species',
                    data_object.taxon_id
                ),
                'attribute_key': self._data_object_factory(
                    'attribute_key',
                    data_object.attribute_key
                ),
                'user': self.__user
            }
        )



# SPDX-FileCopyrightText: 2023 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

import os
import typing
from itertools import chain
from typing import Any, Callable, Iterable, List, Optional

from benchling_sdk.auth.api_key_auth import ApiKeyAuth
from benchling_sdk.benchling import Benchling
from benchling_sdk.errors import BenchlingError, WaitForTaskExpiredError
from benchling_sdk.helpers.retry_helpers import RetryStrategy
from benchling_sdk.models import AsyncTask, AsyncTaskLink, EntityArchiveReason

from cachetools.func import ttl_cache

from caseconverter import snakecase

from more_itertools import batched

from .benchling_converter import (
    BenchlingConverter,
    BenchlingReturn,
    BenchlingWrite,
    DataObjectConverter
)
from ..core import (
    DataObject,
    DataSource,
    DataSourceConfig,
    DataSourceError,
    DataSourceFilter,
    ErrorObject
)
from ..core.operator import (
    Deleter,
    DetailGetter,
    Inserter,
    ListGetter,
    Relational,
    Updater
)
from ..core.operator.updater import DataObjectUpdate
from ..core.relationship import RelationshipConfig

if typing.TYPE_CHECKING:
    from ..core.session import OperableSession

TYPE_MAPPING = {
    'text': 'str',
    'integer': 'int',
    'date': 'datetime',
    'float': 'float',
    'dropdown': 'str',
    'storage_link': 'str',
    'blob_link': 'str',
    'dna_sequence_link': 'str'
}
NATIVE_OBJECT_TYPES = {
    'folder': {
        'attributes': {'name': 'str'},
        'to_one': {'parent_folder': 'folder'}
    },
    'worklist': {
        'attributes': {'name': 'str', 'worklist_type': 'str'}
    },
    'worklist_item': {
        'attributes': {'name': 'str'}
    },
    'transfer': {
        'attributes': {
            'source_entity_id': 'str',
            'destination_container_id': 'str'
        }
    },
    'assay_result': {
        'attributes': {}
    },
    'container_content': {
        'attributes': {
            'batch': 'str',
            'concentration': 'dict',
        },
        'to_one': {
            'entity': 'custom_entity',
        }
    }
}
BENCHLING_TYPE_SEARCH_WITH_SCHEMA_ID = [
    'custom_entity',
    'assay_result',
    'container',
    'plate',
    'box'
]

BENCHLING_PARENT_TYPES_WITH_SCHEMAS = {
    'custom_entity': {
        'attributes': {},
        'to_one': {},
        'to_one_native': {'folder': 'folder'},
        'to_many': {'container_contents': 'container_content'}
    },
    'location': {
        'attributes': {'name': 'str', 'barcode': 'str'},
        'to_one': {'parent_location': 'location'},
        'to_one_native': {},
        'to_many': {}
    },
    'assay_result': {
        'attributes': {},
        'to_one': {},
        'to_one_native': {},
        'to_many': {}
    },
    'container': {
        'attributes': {'barcode': 'str', 'parent_storage_id': 'str'},
        'to_one': {},
        'to_one_native': {},
        'to_many': {'container_contents': 'container_contents'}
    },
    'box': {
        'attributes': {'barcode': 'str'},
        'to_one': {'parent_storage_id': 'location'},
        'to_one_native': {},
        'to_many': {}
    },
    'plate': {
        'attributes': {'barcode': 'str'},
        'to_one': {'parent_storage_id': 'location'},
        'to_one_native': {},
        'to_many': {}
    },
}

BenchlingConverterFactory = Callable[['BenchlingDataSource'], BenchlingConverter]
"""A type hint for the kwarg to `BenchlingDataSource`. Internally, there are no arguments."""
DataObjectConverterFactory = Callable[['BenchlingDataSource'], DataObjectConverter]
"""A type hint for the kwarg to `BenchlingDataSource`. Internally, there are no arguments."""


class BenchlingDataSource(
    DataSource,
    Deleter,
    DetailGetter,
    Inserter,
    ListGetter,
    Relational,
    Updater
):
    """
    A DataSource for writing objects to Benchling
    The queries are maintained in this SDK as SQL files
    """

    url: str
    api_key: str
    registry_id: str
    project_id: str

    def __init__(
        self,
        config: DataSourceConfig,
        benchling_converter_factory: BenchlingConverterFactory | None = None,
        data_object_converter_factory: DataObjectConverterFactory | None = None
    ) -> None:

        # initialy set to `None`
        self.__folder_id = None

        super().__init__(
            config,
            expected=[
                'url',
                'api_key',
                'registry_id',
                'project_id'
            ]
        )

        self.__init_factories(
            benchling_converter_factory,
            data_object_converter_factory,
        )

        self.benchling_interface = self._get_benchling_interface(self.url, self.api_key)
        self.schemas = {
            benchling_type: self._get_schemas(benchling_type)
            for benchling_type in BENCHLING_PARENT_TYPES_WITH_SCHEMAS.keys()
        }

    @property
    def folder_id(self) -> str:
        """The current `folder_id` in Benchling"""

        return (
            self.__folder_id if
            self.__folder_id else
            os.getenv('BENCHLING_FOLDER')
        )

    @folder_id.setter
    def folder_id(self, new_id: str) -> None:
        self.__folder_id = new_id

    @folder_id.deleter
    def folder_id(self) -> None:
        self.__folder_id = None

    def _get_benchling_interface(self, url, api_key):
        return Benchling(
            url=url,
            auth_method=ApiKeyAuth(api_key),
            retry_strategy=RetryStrategy(
                max_tries=3,
                backoff_factor=60.0
            )
        )

    def _get_schemas(
        self,
        benchling_type: str = 'custom_entity'
    ) -> dict[str, dict[str, dict[str, Any]]]:

        pages = self.__get_benchling_schema_function(benchling_type)()
        entities = {}
        for page in pages:
            for schema in page:
                if (
                    (
                        'assay_result' == benchling_type
                        or schema.registry_id == self.registry_id
                    )
                    and schema.archive_record is None
                ):
                    schema_name = snakecase(schema.name)
                    entities[schema_name] = {
                        '__id__': schema.id
                    }
                    for field in schema.field_definitions:
                        if field.archive_record is None:
                            entities[schema_name][snakecase(field.name)] = {
                                'name': field.name,
                                'type': TYPE_MAPPING.get(field.type.value, 'str'),
                                'benchling_type': field.type.value,
                                'required': field.is_required,
                                'is_multi': field.is_multi
                            }
                            if field.type.value == 'dropdown':
                                entities[schema_name][snakecase(field.name)]['dropdown_id'] = \
                                    field.additional_properties.get('dropdownId')
                            if field.type.value == 'entity_link':
                                entities[schema_name][snakecase(field.name)]['schema_id'] = \
                                    field.additional_properties.get('schemaId')
        return entities

    def __get_benchling_package(self, object_type: str):
        if object_type in self.schemas['location'].keys():
            return self.benchling_interface.locations
        elif object_type in self.schemas['box'].keys():
            return self.benchling_interface.boxes
        elif object_type in self.schemas['plate'].keys():
            return self.benchling_interface.plates
        elif object_type in self.schemas['container'].keys():
            return self.benchling_interface.containers
        elif 'transfer' == object_type:
            return self.benchling_interface.containers
        else:
            match self.benchling_types[object_type]: # noqa E999
                case 'folder':
                    return self.benchling_interface.folders
                case 'worklist':
                    return self.benchling_interface.v2.beta.worklists
                case 'location':
                    return self.benchling_interface.locations
                case 'assay_result':
                    return self.benchling_interface.assay_results
                case _:
                    return self.benchling_interface.custom_entities

    def __get_benchling_schema_function(self, benchling_type: str):
        if benchling_type == 'custom_entity':
            return self.benchling_interface.schemas.list_entity_schemas
        function_name = f'list_{benchling_type}_schemas'
        func = getattr(self.benchling_interface.schemas, function_name)
        return func

    @ttl_cache(ttl=86400)
    def get_attribute_value_options(self, object_type: str, name: str) -> dict[str, str]:
        benchling_type = self.benchling_types[object_type]
        dropdown_id = self.schemas[benchling_type][object_type][name]['dropdown_id']
        return {
            option.id: option.name
            for option in self.benchling_interface.dropdowns.get_by_id(dropdown_id).options
        }

    def update(
        self,
        object_type: str,
        updates: Iterable[DataObjectUpdate],
        **kwargs
    ) -> list[DataObject | ErrorObject]:
        """
            Update function of the benchling datasource

            Raises:
                Exception if the object_type is of the assay_result benchling_type.
                This is because the assay result does not support the update method
        """

        if 'assay_result' == self.benchling_types[object_type]:
            raise Exception('Update is not supported on the assay_result benchling_type')

        converter = self.__dc_factory()
        back_converter = self.__bc_factory()

        benchling_package = self.__get_benchling_package(object_type)
        if hasattr(benchling_package, 'bulk_update'):
            return self.__do_bulk_method(
                object_type,
                updates,
                converter,
                back_converter,
                benchling_package.bulk_update,
                benchling_package.update
            )
        else:
            return [
                self.__do_single_method(
                    object_type,
                    update,
                    converter,
                    back_converter,
                    benchling_package.update
                )
                for update in updates
            ]

    def get_by_id(
        self,
        object_type: str,
        object_ids: Iterable[str],
        session=None,
        **kwargs,
    ) -> Iterable[DataObject | ErrorObject | None]:
        back_converter = self.__bc_factory()
        benchling_package = self.__get_benchling_package(object_type)
        benchling_type = self.benchling_types[object_type]
        try:
            kwargs = {}
            if benchling_type in BENCHLING_TYPE_SEARCH_WITH_SCHEMA_ID:
                kwargs['schema_id'] = self.schema_ids[object_type]
            benchling_objects_page = benchling_package.list(
                ids=object_ids,
                **kwargs
            )
            for benchling_objects in benchling_objects_page:
                yield from self.sort_by_id(
                    back_converter.convert_iterable(benchling_objects),
                    object_ids
                )
        except BenchlingError:
            # Fall back to doing a one-by-one lookup
            for object_id in object_ids:
                benchling_object = self.__get_one_by_id(
                    object_type,
                    object_id
                )
                yield back_converter.convert(benchling_object) \
                    if benchling_object is not None else None

    def get_list(
        self,
        object_type: str,
        object_filters: Optional[DataSourceFilter] = None,
        session: Optional[OperableSession] = None,
        **kwargs,
    ) -> Iterable[DataObject]:
        # Currently only deals with filtering by eq/contains: name
        benchling_package = self.__get_benchling_package(object_type)
        benchling_type = self.benchling_types[object_type]
        kwargs = {}
        if object_filters is not None and object_filters.and_ is not None:
            if 'name' in object_filters.and_ and object_filters.and_['name'] is not None:
                if (
                        'contains' in object_filters.and_['name']
                        and object_filters.and_['name']['contains'] is not None
                        and 'value' in object_filters.and_['name']['contains']
                ):
                    kwargs['name_includes'] = object_filters.and_['name']['contains']['value']
                elif (
                    'eq' in object_filters.and_['name']
                    and object_filters.and_['name']['eq'] is not None
                    and 'value' in object_filters.and_['name']['eq']
                ):
                    kwargs['name'] = object_filters.and_['name']['eq']['value']

            if (
                'schema_fields' in object_filters.and_
                and object_filters.and_['schema_fields'] is not None
            ):
                kwargs['schema_fields'] = self.__build_schema_fields_filter(
                    object_filters.and_['schema_fields'],
                    object_type,
                    benchling_type
                )

            for key in ('barcodes', 'entity_ids'):
                if key in object_filters.and_ and object_filters.and_[key] is not None:
                    kwargs[key] = self.__build_list_filter(object_filters.and_[key])

        if benchling_type in BENCHLING_TYPE_SEARCH_WITH_SCHEMA_ID:
            kwargs['schema_id'] = self.schema_ids[object_type]
        # Limit folder searching to the project set a top level
        if object_type == 'folder':
            kwargs['project_id'] = self.project_id

        back_converter = self.__bc_factory()

        try:
            benchling_objects_page = benchling_package.list(
                **kwargs
            )

            for benchling_objects in benchling_objects_page:
                yield from back_converter.convert_iterable(benchling_objects)
        except Exception as e:
            print(f'Exception type: {type(e)}')
            print(f'Exception message: {str(e)}')

    def delete(
        self,
        object_type: str,
        object_ids: Iterable[str],
        session=None
    ) -> None:
        benchling_package = self.__get_benchling_package(object_type)

        if hasattr(benchling_package, 'archive'):
            if (
                object_type in self.schemas['box'].keys()
                or object_type in self.schemas['plate'].keys()
                or object_type in self.schemas['container'].keys()
            ):
                benchling_package.archive(
                    object_ids,
                    reason=EntityArchiveReason.OTHER,  # may need to change this
                    should_remove_barcodes=True
                )
            elif object_type in self.schemas['assay_result'].keys():
                benchling_package.archive(
                    object_ids
                )
            else:
                benchling_package.archive(
                    object_ids,
                    reason=EntityArchiveReason.OTHER  # may need to change this
                )
        elif hasattr(benchling_package, 'delete'):
            for object_id in object_ids:
                benchling_package.delete(object_id)
        else:
            raise DataSourceError(f'Cannot delete {object_type}')

    def insert(
        self,
        object_type: str,
        objects: Iterable[DataObject],
        session=None,
        **kwargs
    ) -> list[DataObject | ErrorObject]:
        converter = self.__dc_factory()
        back_converter = self.__bc_factory()
        benchling_package = self.__get_benchling_package(object_type)

        if object_type == 'worklist_item':
            # Worklist items are appended to a worklist
            return self.__insert_worklist_items(
                objects,
                converter,
                back_converter
            )

        bulk_create_method = None
        create_method = benchling_package.create
        if benchling_package == self.benchling_interface.assay_results:
            bulk_create_method = benchling_package.create
        elif 'transfer' == object_type:
            bulk_create_method = benchling_package.transfer_into_containers
            create_method = benchling_package.transfer_into_containers
        elif hasattr(benchling_package, 'bulk_create'):
            bulk_create_method = benchling_package.bulk_create

        if bulk_create_method is not None:
            # Do bulk inserts of object that allow it
            return self.__do_bulk_method(
                object_type,
                objects,
                converter,
                back_converter,
                bulk_create_method,
                create_method
            )
        else:
            # Other objects are inserted one-by-one
            return [
                self.__do_single_method(
                    object_type,
                    obj,
                    converter,
                    back_converter,
                    create_method
                )
                for obj in objects
            ]

    def __insert_worklist_items(
        self,
        objects: Iterable[DataObject],
        converter: DataObjectConverter,
        back_converter: BenchlingConverter
    ) -> list[DataObject | ErrorObject]:
        """
        Inserts worklist items into a worklist
        """
        for obj in objects:
            worklist_id = obj.worklist.id
            worklist_item_converted = converter.convert(obj)
            try:
                ret = self.__get_benchling_package('worklist').append_item(
                    worklist_id,
                    worklist_item_converted
                )
                yield back_converter.convert(ret)
            except BenchlingError as error:
                yield ErrorObject(
                    error.json['error']['message'],
                    'worklist_item',
                    http_code=error.status_code,
                    object_=obj
                )

    def __do_single_method(
        self,
        object_type: str,
        obj: DataObject,
        converter: DataObjectConverter,
        back_converter: BenchlingConverter,
        method: Callable[[BenchlingWrite], AsyncTaskLink]
    ) -> DataObject | ErrorObject:
        """
        Single object method
        """
        try:
            if isinstance(obj, DataObject):
                converted_object = converter.convert(obj)
                ret = method(converted_object)
            else:
                converted_object = converter.convert_update(obj, object_type)
                # Need to pass in IDs for updates
                ret = method(obj[0], converted_object)
            return back_converter.convert(ret)
        except BenchlingError as error:
            return ErrorObject(
                error.json['error']['message'],
                object_type,
                http_code=error.status_code,
                object_=obj
            )

    def __do_bulk_method(
        self,
        object_type: str,
        objects: Iterable[DataObject],
        converter: DataObjectConverter,
        back_converter: BenchlingConverter,
        bulk_method: Callable[[Iterable[BenchlingWrite]], AsyncTaskLink] = None,
        single_method: Callable[[Iterable[BenchlingWrite]], AsyncTaskLink] = None
    ) -> list[DataObject | ErrorObject]:
        """
        Splits a (potentially long) `Iterable[DataObject]` into smaller
        pages lazily, and calls the given `bulk_method` (with retries)
        on each, chaining the results together.
        """
        batched_pages = batched(objects, 20)

        output_pages = (
            self.__do_bulk_method_on_page(
                object_type,
                list(page),
                converter,
                back_converter,
                bulk_method,
                single_method
            )
            for page in batched_pages
        )

        return list(
            chain.from_iterable(output_pages)
        )

    def __do_bulk_method_on_page(
        self,
        object_type: str,
        page: list[DataObject],
        converter,
        back_converter: BenchlingConverter,
        bulk_method: Callable[[Iterable[BenchlingWrite]], AsyncTaskLink],
        single_method: Callable[[Iterable[BenchlingWrite]], AsyncTaskLink],
    ) -> Iterable[DataObject | ErrorObject]:
        """
        Calls the given `bulk_method` on the `page` of `DataObject`
        instances.

        If this page fails, it is retried with each element
        individually.
        """

        if isinstance(page[0], DataObject):
            converted_objects = converter.convert_bulk(page, object_type)
        else:
            converted_objects = converter.convert_bulk_update(page, object_type)

        if 'assay_result' == self.benchling_types[object_type]:
            response = bulk_method(converted_objects)

            if hasattr(response, '_errors'):
                return self.__retry_bulk_methods_on_singletons(
                    object_type,
                    page,
                    converter,
                    back_converter,
                    single_method
                )

            return_objects = [response]
        else:
            task = self.__wait_for_task(
                converted_objects,
                bulk_method
            )

            if task.status == 'FAILED':
                return self.__retry_bulk_methods_on_singletons(
                    object_type,
                    page,
                    converter,
                    back_converter,
                    single_method
                )

            return_key = 'customEntities'
            if object_type in self.schemas['container'].keys():
                return_key = 'containers'
            elif 'transfer' == object_type:
                return_key = 'destinationContainers'

            return_objects = task.response.additional_properties[return_key]
        return back_converter.convert_return_entities(return_objects)

    def __retry_bulk_methods_on_singletons(
        self,
        object_type: str,
        page: list[DataObject],
        converter: DataObjectConverter,
        back_converter: BenchlingConverter,
        single_method: Callable[[Iterable[BenchlingWrite]], AsyncTaskLink],
    ) -> list[DataObject | ErrorObject]:
        """
        Retries a failed page, iterating the bulk method
        on a singleton of each element.
        """
        for obj in page:
            yield self.__do_single_method(
                object_type,
                obj,
                converter,
                back_converter,
                single_method
            )

    def __wait_for_task(
        self,
        request: Iterable[BenchlingWrite],
        bulk_method: Callable[[Iterable[BenchlingWrite]], AsyncTaskLink],
    ) -> AsyncTask:

        try:
            response = bulk_method(request)
            return self.benchling_interface.tasks.wait_for_task(
                response.task_id,
                interval_wait_seconds=5
            )
        except WaitForTaskExpiredError:
            raise DataSourceError(
                'Time out in communication with Benchling '
                '(waiting for task response).',
                '',
                status_code=400
            )
        except BenchlingError as error:
            raise DataSourceError(
                'Error creating update task',
                error.json['error']['message'],
                status_code=400
            )

    @property
    def attribute_types(self):
        return {
            k1: {
                k2: v2['type']
                for k2, v2 in v1.items()
                if not k2.startswith('__')  # filter out '__id__'
                and v2['benchling_type'] != 'entity_link'
            } | BENCHLING_PARENT_TYPES_WITH_SCHEMAS[benchling_type]['attributes']
            for benchling_type in BENCHLING_PARENT_TYPES_WITH_SCHEMAS.keys()
            for k1, v1 in self.schemas[benchling_type].items()
        } | {
            k: v['attributes']
            for k, v in NATIVE_OBJECT_TYPES.items()
        }

    @property
    def supported_types(self) -> List:
        result = [
            key
            for benchling_type in BENCHLING_PARENT_TYPES_WITH_SCHEMAS.keys()
            for key in self.schemas[benchling_type].keys()
        ]
        result.extend(list(NATIVE_OBJECT_TYPES.keys()))
        return result

    @property
    def schema_ids(self) -> dict[str, str]:
        """
        Maps elements of `supported_types` to their
        `schema_id` in benchling.
        """

        return {
            k: v['__id__']
            for benchling_type in BENCHLING_PARENT_TYPES_WITH_SCHEMAS.keys()
            for k, v
            in self.schemas[benchling_type].items()
        }

    @property
    def schema_names(self) -> dict[str, str]:
        return {
            v['__id__']: k
            for benchling_type in BENCHLING_PARENT_TYPES_WITH_SCHEMAS.keys()
            for k, v
            in self.schemas[benchling_type].items()
        }

    @property
    def benchling_types(self) -> dict[str, str]:
        return {
            k: benchling_type
            for benchling_type in BENCHLING_PARENT_TYPES_WITH_SCHEMAS.keys()
            for k in self.schemas[benchling_type].keys()
        } | {k: k for k in NATIVE_OBJECT_TYPES.keys()}

    def get_page_size(self) -> int:
        return 20

    def __get_one_by_id(
        self,
        object_type: str,
        object_id: str
    ) -> BenchlingReturn:

        benchling_package = self.__get_benchling_package(object_type)
        benchling_type = self.benchling_types[object_type]
        try:
            if benchling_type in BENCHLING_TYPE_SEARCH_WITH_SCHEMA_ID:
                obj = benchling_package.get_by_id(
                    object_id,
                )
                if snakecase(obj.schema.name) == object_type:
                    return obj
                return None

            return benchling_package.get_by_id(
                object_id
            )
        except BenchlingError:
            return None

    def __init_factories(
        self,
        benchling_converter_factory: BenchlingConverterFactory | None,
        data_object_converter_factory: DataObjectConverterFactory | None
    ) -> None:

        self.__bc_factory = (
            self.__default_bc_factory
            if benchling_converter_factory is None
            else lambda: benchling_converter_factory(self)
        )

        self.__dc_factory = (
            self.__default_dc_factory
            if data_object_converter_factory is None
            else lambda: data_object_converter_factory(self)
        )

    def __default_bc_factory(self) -> BenchlingConverter:
        return BenchlingConverter(self)

    def __default_dc_factory(self) -> DataObjectConverter:
        return DataObjectConverter(self)

    @property
    def relationship_config(self) -> dict[str, RelationshipConfig]:
        """
        The configuration of relationships (both to-one and to-many) between
        the types of DataObject instances managed by this DataSource instance.
        """
        return {
            'folder': RelationshipConfig(
                to_one={
                    'parent_folder': 'folder'
                },
                to_many={}
            ),
            'worklist': RelationshipConfig(
                to_many={
                    'worklist_items': 'worklist_item'
                }
            ),
            'worklist_item': RelationshipConfig(
                to_one={
                    'worklist': 'worklist',
                    'item': list(self.schemas['custom_entity'].keys())
                    + list(self.schemas['container'].keys()),
                }
            ),
            'container_content': RelationshipConfig(
                to_one={
                    'entity': list(self.schemas['custom_entity'].keys()),
                    'container': list(self.schemas['container'].keys()),
                },
                to_many={}
            )
        } | {
            k: RelationshipConfig(
                to_one={
                    attribute_name: self.schema_names.get(entity_def['schema_id'])
                    for attribute_name, entity_def in v.items()
                    if 'schema_id' in entity_def and entity_def['is_multi'] is False
                } | {
                    k1: list(self.schemas[v1].keys())
                    for k1, v1 in BENCHLING_PARENT_TYPES_WITH_SCHEMAS[benchling_type][
                        'to_one'
                    ].items()
                } | BENCHLING_PARENT_TYPES_WITH_SCHEMAS[benchling_type]['to_one_native'],
                to_many=BENCHLING_PARENT_TYPES_WITH_SCHEMAS[benchling_type]['to_many']
            )
            for benchling_type in BENCHLING_PARENT_TYPES_WITH_SCHEMAS.keys()
            for k, v in self.schemas[benchling_type].items()
        }

    def get_to_one_relation(
        self,
        source: DataObject,
        relationship_name: str,
        session: Optional[OperableSession] = None
    ) -> Optional[DataObject]:
        """
        Gets the to-one relation DataObject, given a source DataObject and the
        name of the relationship within the config.
        """
        pass

    def get_to_many_relations(
        self,
        source: DataObject,
        relationship_name: str,
        session: Optional[OperableSession] = None
    ) -> Iterable[DataObject]:
        """
        Gets the Iterable of to-many relation DataObject instances, given a source
        DataObject and the name of the relationship within the config.
        """
        if source.type == 'worklist' and relationship_name == 'worklist_items':
            back_converter = self.__bc_factory()
            yield from back_converter.convert_worklist_items(
                self.__get_benchling_package('worklist').get_by_id(source.id)
            )

        if (
            source.type in self.schemas['container'].keys()
            and relationship_name == 'container_contents'
        ):
            back_converter = self.__bc_factory()
            contents = self.get_container_contents(source.id)
            yield back_converter.convert_container_contents(contents)

        if (
            source.type in self.schemas['custom_entity'].keys()
            and relationship_name == 'container_contents'
        ):
            back_converter = self.__bc_factory()
            container_list = self.benchling_interface.containers.list(
                storage_contents_id=source.id
            )
            for container in container_list:
                for individual_container in container:
                    converted_container = back_converter.convert(individual_container)
                    contents = self.get_container_contents(converted_container.id)
                    converted_contents = back_converter.convert_container_contents(contents)
                    for content in converted_contents:
                        if content.to_one_relationships['entity'].id == source.id:
                            yield content

    def __build_list_filter(self, list_filter: Optional[DataSourceFilter]):
        arg = ''

        if (
            'eq' in list_filter
            and list_filter['eq'] is not None
            and 'value' in list_filter['eq']
            and list_filter['eq']['value'] is not None
        ):
            arg = [list_filter['eq']['value']]
        elif (
            'in_list' in list_filter
            and list_filter['in_list'] is not None
            and 'value' in list_filter['in_list']
            and list_filter['in_list']['value'] is not None
            and isinstance(list_filter['in_list']['value'], list)
        ):
            arg = list_filter['in_list']['value']

        return arg

    def __build_schema_fields_filter(
        self,
        schema_filters: Optional[DataSourceFilter],
        object_type: str,
        benchling_type: str
    ):
        kwargs = {}

        if self.attribute_types.get(object_type):
            fields = self.schemas[benchling_type][object_type]

            for key, field_data in fields.items():
                filter_criteria = schema_filters.and_.get(key)

                if not filter_criteria:
                    continue

                field_type = field_data['type']

                # benchling expexts the field name as the key for the filter
                field_key = field_data['name']

                eq_value = filter_criteria.get('eq', {}).get('value')
                if field_type == 'str' and eq_value is not None:
                    benchling_data_type = field_data['benchling_type']

                    if benchling_data_type == 'dropdown':
                        options = self.get_attribute_value_options(object_type, key)
                        for identifier, option in options.items():
                            if option == eq_value:
                                eq_value = identifier
                                break

                    kwargs[field_key] = eq_value
                elif field_type in ['int', 'datetime', 'float']:
                    if eq_value is not None:
                        kwargs[field_key] = eq_value
                    else:
                        gt_eq_value = filter_criteria.get('gt_eq', {}).get('value')
                        ls_eq_value = filter_criteria.get('ls_eq', {}).get('value')
                        between_values = filter_criteria.get('between', {})

                        if gt_eq_value:
                            kwargs[field_key] = f'>={gt_eq_value}'
                        elif ls_eq_value:
                            kwargs[field_key] = f'<={ls_eq_value}'
                        elif (
                            'value_smaller' in between_values
                            and 'value_larger' in between_values
                        ):
                            kwargs[field_key] = (
                                f">={between_values['value_smaller']}"
                                f"<={between_values['value_larger']}"
                            )

        return kwargs

    def __clean_object_type(self, object_type):
        if object_type in self.benchling_types:
            return self.benchling_types[object_type]

        return object_type

    def get_container_contents(self, container_id):
        container_package = self.benchling_interface.containers

        return container_package.list_contents(container_id)



# SPDX-FileCopyrightText: 2024 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from ..core import (
    core_data_object
)
from ..json import (
    JsonDataSource
)


def tolqc_legacy(**kwargs) -> JsonDataSource:
    tolqc_legacy = JsonDataSource({
        'uri': 'https://tolqc.cog.sanger.ac.uk/data.json',
        'type': 'species',
        'id_attribute': 'taxon',
        'mappings': {
            'group': {
                'heading': 'group',
                'type': 'str'
            },
            'hierarchy_name': {
                'heading': '_name',
                'type': 'str'
            },
            'assembly_stage': {
                'heading': 'asm.stage',
                'type': 'str'
            }
        }
    })
    core_data_object(tolqc_legacy)
    return tolqc_legacy



# SPDX-FileCopyrightText: 2024 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from abc import ABC


def get_operator_member_names(
    class_: type[ABC]
) -> list[str]:
    """
    Gets a list of all operator methods
    """

    return [
        k
        for k in vars(class_).keys()
        if not k.startswith('_')
    ]



# SPDX-FileCopyrightText: 2023 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

import glob

from flask import Blueprint


def system_blueprint(
    url_prefix: str = '/system',
    env_map: dict[str, str] = {'environment': 'ENVIRONMENT'},
    env_vars: dict[str, str] = {}
) -> Blueprint:
    """
    A flask Blueprint that groups system endpoints.

    By default includes:
    - /healthz      - used for health probes in kubernetes
    - /environment  - indicates the deployment environment
    - /package-versions  - indicates the deployment package versions

    Parameters:
    - url_prefix    - the prefix under which to serve this blueprint's
                      endpoints
    - env_map       - maps keys in the /environment output to keys in
                      the env_vars dict
    - env_vars      - a dict containing environment variables (defaults
                      to os.environ)
    """

    system_blueprint = Blueprint('system', __name__, url_prefix=url_prefix)

    @system_blueprint.route('/healthz')
    def healthz():
        return {
            'healthy': True
        }, 200

    @system_blueprint.route('/environment')
    def environment():
        return {
            k: env_vars.get(v) for k, v in env_map.items()
        }, 200

    @system_blueprint.route('/package-versions')
    def package_versions():
        glob_paths = glob.glob('./**/requirements.txt', recursive=True)
        with open(glob_paths[0], 'r') as f:
            requirements = f.readlines()
            for requirement in requirements:
                if 'tol-sdk' in requirement:
                    tol_sdk_version = requirement.split('==')[1].strip()
                    break
            f.close()

        result = {'tol_sdk': tol_sdk_version}

        return result, 200

    return system_blueprint



# SPDX-FileCopyrightText: 2025 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from dataclasses import dataclass
from typing import Any, List

from tol.core import DataObject, Validator

from .interfaces import Condition, ConditionDict, ConditionEvaluator


class MutuallyExclusiveValidator(Validator, ConditionEvaluator):
    """
    Validates an incoming stream of `DataObject` instances,
    where the resultant field from field_one_condition must not
    have the same values for target_fields as the resultant
    field from field_two_condition
    """
    @dataclass(slots=True, frozen=True, kw_only=True)
    class Config:
        first_field_where: ConditionDict
        second_field_where: ConditionDict
        target_fields: List[str]
        detail: str | None = None

        def _get_error_message(self) -> str:
            # Extract conditions
            first_condition = Condition.from_dict(self.first_field_where)
            second_condition = Condition.from_dict(self.second_field_where)

            # Use a pre-defined, hard-coded detail message if one was not provided
            if self.detail is None:
                multiple_target_fields = len(self.target_fields) > 1
                possible_plural = 's' if multiple_target_fields else ''

                target_fields_str = ''
                if multiple_target_fields:
                    for index, field in enumerate(self.target_fields):
                        if index == 0:
                            # First field in the list
                            target_fields_str += f'{field}'
                        elif index == len(self.target_fields) - 1:
                            # Last field in the list
                            target_fields_str += f' and {field}'
                        else:
                            # Middle fields
                            target_fields_str += f', {field}'
                else:  # Only one field
                    target_fields_str = self.target_fields[0]

                return (
                    f'The field{possible_plural} {target_fields_str} cannot have the same '
                    f'value{possible_plural} both when {first_condition} and when '
                    f'{second_condition}'
                )
            else:
                return self.detail

    __slots__ = ['__config', '__first_list', '__second_list']
    __config: Config
    __first_list: List[Any]
    __second_list: List[Any]

    def __init__(self, config: Config, **kwargs) -> None:
        super().__init__()

        self.__config = config
        self.__first_list = []
        self.__second_list = []

    def _validate_data_object(self, obj: DataObject) -> None:
        # Check first field
        if self._does_condition_pass(Condition.from_dict(self.__config.first_field_where), obj):
            # Check whether the values of the target fields were found in the second list
            if [
                obj.get_field_by_name(target_field)
                for target_field in self.__config.target_fields
            ] in self.__second_list:
                self.add_error(
                    object_id=obj.id,
                    detail=self.__config._get_error_message()
                )

            # Add the values of the target fields to the first list
            self.__first_list.append(
                [
                    obj.get_field_by_name(target_field)
                    for target_field in self.__config.target_fields
                ]
            )
        # Check second field (same as the first condition, but for the second!)
        elif self._does_condition_pass(Condition.from_dict(self.__config.second_field_where), obj):
            # Check whether the values of the target fields were found in the first list
            if [
                obj.get_field_by_name(target_field)
                for target_field in self.__config.target_fields
            ] in self.__first_list:
                self.add_error(
                    object_id=obj.id,
                    detail=self.__config._get_error_message()
                )

            # Add the values of the target fields to the second list
            self.__second_list.append(
                [
                    obj.get_field_by_name(target_field)
                    for target_field in self.__config.target_fields
                ]
            )
        # If neither condition passes, the data object must be valid (for this validator anyway!)



# SPDX-FileCopyrightText: 2023 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from typing import Callable, Dict, List, Type

from .database import Database, DefaultDatabase
from .filter import DefaultDatabaseFilter
from .model import Model
from .relationship import DefaultSqlRelationshipConfig
from .session import create_session_factory
from .sort import DefaultDatabaseSorter
from .sql_converter import DefaultDataObjectConverter, DefaultModelConverter, TypeFunction
from .sql_datasource import (
    BackConverterFactory,
    ConverterFactory,
    FilterFactory,
    SorterFactory,
    SqlDataSource
)
from ..api_base.misc.auth_context import default_ctx_getter


DatabaseFactory = Callable[
    [List[Type[Model]], str],
    Database
]


def __model_converter_factory(
    type_function: TypeFunction
) -> ConverterFactory:

    return lambda do_factory, req_fields_tree: DefaultModelConverter(
        type_function,
        do_factory,
        requested_tree=req_fields_tree
    )


def __back_converter_factory(
    models: List[Type[Model]],
    type_function: TypeFunction
) -> BackConverterFactory:

    models_dict = {
        type_function(m): m for m in models
    }
    return lambda: DefaultDataObjectConverter(
        models_dict
    )


def __sorter_factory() -> SorterFactory:
    return lambda sort_term: DefaultDatabaseSorter(
        sort_term
    )


def __type_tablename_dict(
    models: List[Type[Model]],
    type_function: TypeFunction
) -> Dict[str, str]:
    """Inverts the TypeFunction callable, using a dictionary"""

    return {
        type_function(m): m.get_table_name() for m in models
    }


def __filter_factory() -> FilterFactory:

    return lambda ds_filter: DefaultDatabaseFilter(
        ds_filter,
    )


def __database(
    models: List[Type[Model]],
    db_uri: str
) -> Database:

    session_factory = create_session_factory(db_uri)
    return DefaultDatabase(session_factory, models)


def create_sql_datasource(
    models: List[Type[Model]],
    db_uri: str,
    behind_api: bool = False,

    # arguments below are optional and used mainly for testing
    type_function: TypeFunction = lambda m: m.get_table_name(),
    api_user_id_getter: Callable[[], str] = lambda: default_ctx_getter().user_id,
    database_factory: DatabaseFactory = __database,
    model_factory: BackConverterFactory = __back_converter_factory,
    do_factory: ConverterFactory = __model_converter_factory,
    **kwargs
) -> SqlDataSource:
    """
    Creates an SqlDataSource instance using:

    - a list of Model classes
    - a string database URI
    - an (optional) callable that gets the DataObject type for a given Model class
    """

    converter_factory = do_factory(type_function)
    back_converter_factory = model_factory(models, type_function)
    sorter_factory = __sorter_factory()
    type_tablename_dict = __type_tablename_dict(models, type_function)
    sql_relationship_config = DefaultSqlRelationshipConfig(models, type_function)
    filter_factory = __filter_factory()
    db = database_factory(models, db_uri)

    user_id_getter = api_user_id_getter if behind_api else None

    return SqlDataSource(
        db,
        type_tablename_dict,
        sql_relationship_config,
        converter_factory,
        back_converter_factory,
        filter_factory,
        sorter_factory,
        user_id_getter=user_id_getter,
        **kwargs
    )



# SPDX-FileCopyrightText: 2024 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from typing import Iterable

from ...core import (
    DataObject,
    DataObjectToDataObjectOrUpdateConverter
)


class LabwhereLocationToStsTrayConverter(
        DataObjectToDataObjectOrUpdateConverter):
    def convert(self, data_object: DataObject) -> Iterable[DataObject]:
        if data_object is not None:
            yield self._data_object_factory(
                'freezer_tray',
                data_object.id,
                attributes={
                    'name': data_object.name,
                    'parentage': data_object.parentage
                }
            )



# SPDX-FileCopyrightText: 2025 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from abc import ABC
from dataclasses import dataclass
from typing import Any, Dict, Tuple, cast

from tol.core import DataObject


ConditionDict = Dict[str, str | Any | bool]
"""
The dict representation of a Condition. Conditions can be constructed
from such a dict through Condition.from_dict(condition_dict)
"""


@dataclass(slots=True)
class Condition:
    field: str
    operator: str
    value: Any
    # If this condition fails, should it be an error or a warning?
    # The reporting of this error or warning is done in the calling validator
    is_error: bool = True

    def __repr__(self) -> str:
        return f'{self.field} {self.operator} {self.value}'

    @staticmethod
    def from_dict(condition_dict: ConditionDict) -> 'Condition':
        """
        A means of instantiating a Condition from a dictionary.
        This is a separate method (rather than constructing with kwargs
        like `Condition(**condition_dict))`) to allow for both precense
        and type checking for each field.
        """
        try:
            # Extract fields
            field = condition_dict['field']
            operator = condition_dict['operator']
            value = condition_dict['value']
            is_error = condition_dict.get('is_error', True)

            # Ensure fields are the correct type
            if not isinstance(field, str) and not isinstance(operator, str):
                raise Exception(
                    f'Dictionary {condition_dict} not in valid format '
                    f'to convert to Condition (type of condition dict incorrect)'
                )

            return Condition(
                cast(str, field),
                cast(str, operator),
                value,
                cast(bool, is_error),
            )
        except IndexError as e:
            raise Exception(
                f'Dictionary {condition_dict} not in valid format '
                f'to convert to Condition (key "{e.args[0]}" not found)'
            )


class ConditionEvaluator(ABC):
    """
    Interface to be inherited by validators. Evaluates the provided condition given its
    operator and operands
    """
    def _evaluate_condition(self, condition: Condition, obj: DataObject) -> Tuple[bool, Any]:
        """
        Evaluates the provided condition given its operator and operands.
        If `operator` is not one of the supported operators, an exception is thrown.
        """
        value_to_test = obj.get_field_by_name(condition.field)

        match condition.operator:
            case '==':
                return (value_to_test == condition.value, value_to_test)
            case '!=':
                return (value_to_test != condition.value, value_to_test)
            case '<':
                return (value_to_test < condition.value, value_to_test)
            case '<=':
                return (value_to_test <= condition.value, value_to_test)
            case '>':
                return (value_to_test > condition.value, value_to_test)
            case '>=':
                return (value_to_test >= condition.value, value_to_test)
            case 'in':
                return (value_to_test in condition.value, value_to_test)
            case _:
                raise Exception(f'VALIDATOR SETUP ERROR: `{condition.operator}` is not '
                                f'a supported operator for {type(self).__name__}')

    def _does_condition_pass(self, condition: Condition, obj: DataObject) -> bool:
        """
        Helper function for when you only want to know whether the condition passes,
        and don't need the actual value
        """
        return self._evaluate_condition(condition, obj)[0]



# SPDX-FileCopyrightText: 2023 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

import typing
from abc import ABC, abstractmethod
from typing import Iterable, Optional

if typing.TYPE_CHECKING:
    from ..data_object import DataObject
    from ..requested_fields import ReqFieldsTree
    from ..session import OperableSession

from more_itertools import chunked, seekable


class DetailGetter(ABC):
    """
    Gets an Iterable of (Optional) DataObject instances given an
    Iterable of ID strings.
    """
    page_size = 20

    def get_by_ids(
        self,
        object_type: str,
        object_ids: Iterable[str],
        session: Optional[OperableSession] = None,
        **kwargs,
    ) -> Iterable[Optional[DataObject]]:
        """
        Gets an Iterable of DataObject instances, of specified object_type,
        with their id's equal to those given in the object_ids Iterable (or
        None if the id at that position is not found).
        This splits up the request to get_by_id into sensible size
        batches, so we can safely pass a long list to this method
        """

        for chunk in chunked(object_ids, self.page_size):
            yield from self.get_by_id(object_type, chunk, **kwargs)

    @abstractmethod
    def get_by_id(
        self,
        object_type: str,
        object_ids: Iterable[str],
        session: Optional[OperableSession] = None,
        requested_fields: list[str] | None = None,
        requested_tree: ReqFieldsTree | None = None,
    ) -> Iterable[Optional[DataObject]]:
        """
        Gets an Iterable of DataObject instances, of specified object_type,
        with their id's equal to those given in the object_ids Iterable (or
        None if the id at that position is not found).
        """

    def get_one(
        self,
        object_type: str,
        object_id: str,
        session: Optional[OperableSession] = None,
        **kwargs,
    ) -> Optional[DataObject]:
        """
        Gets the individual `DataObject` instance, of specified object_type
        and object_id, or returns `None` if not found.
        """

        return list(self.get_by_id(object_type, [object_id], **kwargs))[0]

    # A helper method to ensure that the order of the returned objects
    # matches the order of the input ids
    def sort_by_id(
        self,
        data_objects: Iterable[DataObject],
        object_ids: Iterable[int | str]
    ) -> Iterable[DataObject | None]:

        seekable_objects = seekable(data_objects)
        for id_ in object_ids:
            seekable_objects.seek(0)
            for obj in seekable_objects:
                if str(obj.id) == str(id_):
                    yield obj
                    break
            else:
                yield None



# SPDX-FileCopyrightText: 2025 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from tempfile import NamedTemporaryFile
from typing import Any, Protocol

import gdown

from .csv_datasource import CsvDataSource


class GoogleFetcher(Protocol):
    def __call__(
        self,
        google_file_id: str,
        local_filename: str,
    ) -> None:
        ...


def fetch_from_google_drive(
    google_file_id: str,
    local_filename: str,
) -> None:
    gdown.download(id=google_file_id, output=local_filename)


def google_csv_datasource_factory(
    *,
    google_file_id: str,
    google_fetcher: GoogleFetcher = fetch_from_google_drive,
    **kwargs: Any,
) -> CsvDataSource:

    with NamedTemporaryFile() as temp_fil:
        filename = temp_fil.name

        google_fetcher(google_file_id, filename)

        return CsvDataSource(
            filename,
            **kwargs,
        )
